//////////////////////////////////////////////////////////////////////////////
//
// File:	blind.C
// Directory:	/users/hartmann/rapid/src
//
//////////////////////////////////////////////////////////////////////////////

#include "blind.h"

// key characters
const char	SepEQU  = '=';
const char	SepDDOT = ':';
const char	EndMARK = '#';
const char	LeftSEC = '[';
const char	RightSEC= ']';

// file status flags
const int 	FClosed  = 0;
const int 	FOpenedR = 1;
const int 	FOpenedW = 2;

int errno;

// static declarations needed for gcc compiler

char* 		blindBox::FileName		= 0;
char		blindBox::Buffer[BufferSIZE] 	= {0};
FILE* 		blindBox::File 			= 0;
int 		blindBox::FStatus	 	= 0;
unsigned int 	blindBox::InitFlag 		= 0;
long int 	blindBox::SectionOffset 	= 0;


//// blindBox::blindBox /////////////////////////////////
//

blindBox::blindBox(const char *SignStr, char *FName)
{
    
   if (FName != NULL) 
      if(InitFlag)
      {
	 WarnOutput << Sign 
	    << "Change of filename to '" << FName << "' ignored.";
	 if (FileName != NULL)
	    cerr << "It has been previously set to '" << FileName;
	    cerr << '\n';
      }
      else
      {
	 InitFlag++;
         FileName = FName;
         FStatus = FClosed;
	 SectionOffset = (long) -1;
      }
   
   if(!(Signature = strnewcpy(SignStr)))
      exitus;
   
}


blindBox::blindBox(const char *SignStr)
{
      
   if(!(Signature = strnewcpy(SignStr)))
      exitus;
   
}



//// blindBox::Close /////////////////////////////////
//

void blindBox::Close()
{
 
   if (!InitFlag)
   {
      ErrOutput << Sign 
	 << "File not specified yet. Unable to close it ... .\n";
      exitus;
   }
   
   if (FStatus != FClosed)
   {
      if (fclose(File))
      {
         ErrNOutput(errno) << Sign << "Unable to close '" << FileName << "'\n";
	 exit(errno);
      }
      else
         FStatus = FClosed;
   }

}


//// blindBox::OpenR /////////////////////////////////
//
   
void blindBox::OpenR()
{

   if (FStatus != FOpenedR)
   {
      Close();
      if ((File=fopen(FileName,"r")) == NULL)
      {
         ErrNOutput(errno) << Sign << "Unable to open '" << FileName << "'\n";
	 exit(errno);
      }

      SectionOffset = (long) -1;
      FStatus = FOpenedR;
   }

}


//// blindBox::OpenW /////////////////////////////////
//

void blindBox::OpenW()
{
   
   if (FStatus != FOpenedW)
   {
      Close();
      if ((File=fopen(FileName,"w")) == NULL)
      {
         ErrNOutput(errno) << Sign << "Unable to open '" << FileName << "'\n";
	 exit(errno);
      }

      FStatus = FOpenedW;
   }

}


//// blindBox::Getc /////////////////////////////////
//

char blindBox::Getc()
{

   OpenR();
   return getc(File);

}


//// blindBox::Ungetc /////////////////////////////////
//

void blindBox::Ungetc()
{
   
   OpenR();
   
   if(ftell(File) == 0L)
      return;
   
   if(fseek(File,(long) -1,SEEK_CUR))
   {
      ErrNOutput(errno) << Sign 
	 << "Unable to unget character in the file '" << FileName << "'\n";
      exit(errno) ;
   }
      
}


//// blindBox::Fseek /////////////////////////////////
//

void blindBox::Fseek(long int Offset)
{
   
   OpenR();
   if(Offset > 0)
   {
      if(fseek(File, Offset,SEEK_SET))
      {
	 ErrNOutput(errno) << Sign 
	    << "Unable to set the file position in '" << FileName << "'\n";
         exit(errno);
      }
      else
	 return;
   }
   else
      rewind(File);

}


//// blindBox::Ftell /////////////////////////////////
//

long int blindBox::Ftell()
{
   
   long int Offset;
   
   if((Offset = ftell(File)) < 0L )
   {
      ErrNOutput(errno) << Sign 
	 << "Unable to get the position in '" << FileName << "'\n";
      exit(errno);
   }
   
   return Offset;
   
}
   

//// blindBox::CheckChar /////////////////////////////////
//

boolean blindBox::CheckChar(char X)
{
   
   char c;

   if((c=Getc()) == X)
      return TRUE;

   if(c != EOF)
      Ungetc();

   return FALSE;
   
}


//// blindBox::CheckStr /////////////////////////////////
//

boolean blindBox::CheckStr(char *Str)
{
   
   long int Offset = Ftell();
   char c;

   while(*Str)
      if( (c=Getc()) == *Str)
         Str++;
      else
	 break;
   
   if(*Str)
   {
      Fseek(Offset);

      return FALSE;
   }
   else
      return TRUE;
   
}


//// blindBox::IgnoreSpc /////////////////////////////////
//

void blindBox::IgnoreSpc()
{
   
   char c;
   
   while((c = Getc()) != EOF)
      if(!isspace(c))
      {
	 Ungetc();
	 return;
      }
   
   return;
	 
}


//// blindBox::IgnoreSpcLn /////////////////////////////////
//

void blindBox::IgnoreSpcLn()
{
   
   char c;
   
   while((c = Getc()) != EOF)
      if((!isspace(c)) || (c=='\n'))
      {
	 Ungetc();
	 return;
      }
   
   return;
	 
}
   

//// blindBox::IgnoreSet /////////////////////////////////
//

void blindBox::IgnoreSet(const char *Set)
{
   
   char c;
   const char *p;

   while((c = Getc()) != EOF)
   {
      p = Set;
      while((*p != c) && *p)
         p++;
   
      if(!*p)
      {
	 Ungetc();
	 return;
      }
   }
   
   return;

}


//// blindBox::GotoNextLn /////////////////////////////////
//

void blindBox::GotoNextLn()
{
   
   char c;

   while((c=Getc()) != EOF)
      if(c == '\n')
         return;
   
   return;
   
}


//// blindBox::ReadStrLn /////////////////////////////////
//

void blindBox::ReadStrLn(char *LBuffer)
{
   int i=0;
   char c;
   
   IgnoreSpcLn();
   while((c=Getc()) != EOF)
   {
      if(c == '\n')
	 Ungetc();
      
      if(isspace(c))
         break;
      else
      {
         if(i < (BufferSIZE - 1))
	    LBuffer[i++] = c;
	 else
	 {
	    ErrOutput << Sign << "String '" << LBuffer << "' is too long.\n";
	    exitus;
	 }
       }
   }  
   LBuffer[i] = '\0';

   return;
   
}


//// blindBox::FindSection /////////////////////////////////
//

result blindBox::FindSection(const char *SectionId)
{
   
   if(!SectionId)
      SectionId = Signature;
   
   char c;
   
   OpenR();
   long int Offset = Ftell();
   if(*SectionId)
   {
      for(int i=0; i<2; i++)
      {
	 if(Ftell() > 0L)
	    GotoNextLn();
	 
	 do
	 {
	    IgnoreSpc();
	    if((c=Getc()) == LeftSEC)
	    {
	       IgnoreSpcLn();
	       if(CheckStr((char*)SectionId) == TRUE)
	       {
		  IgnoreSpcLn();
                  if(CheckChar(RightSEC) == TRUE)
	          {
		     GotoNextLn();
		     Ungetc();
		     SectionOffset = Ftell();
		     
		     return OKAY;
	          }
	       }
	    }
 	    GotoNextLn();
	 }
	 while((c!=EOF) && ((i==0) || (i && (Ftell()<Offset))));
	    
	 if(i)
	    Fseek(Offset);
	 else
	    Fseek(0L);
      }
      SectionOffset = (long) -1;
      
      return EMPTY;

   }
   else
   {
      Fseek(SectionOffset=0L);

      return OKAY;
   }

}
   
   
//// blindBox::FindIdentifier /////////////////////////////////
//

result blindBox::FindIdentifier(char *Id, char Separ)
{

   long int Offset = Ftell();
   
   if((*Id) && (SectionOffset > 0L))
   {
      for(int i=0; i<2; i++)
      {
	 if(Ftell() > 0L)
	    GotoNextLn();
	 
	 do
	 {
	    IgnoreSpc();
	    if((CheckChar(LeftSEC) == TRUE) || (CheckChar(EOF) == TRUE))
	       break;
	    else
	       if(CheckStr(Id) == TRUE)
	       {
		  IgnoreSpcLn();
                  if(CheckChar(Separ) == TRUE)

		     return OKAY;
	       }
	    GotoNextLn();
	 }
	 while((i==0) || (i && (Ftell()<Offset)));
	 
	 if(i)
	    Fseek(Offset);
	 else
	    Fseek(SectionOffset);

      }
   }

   return EMPTY;

}


//// blindBox::ReadVal /////////////////////////////////
//

result blindBox::ReadVal()
{
 
   ReadStrLn(Buffer);
   if(*Buffer)
      
      return OKAY;
   
   else
      
      return EMPTY;
   
}


result blindBox::ReadVal(int &Var)
{

   long int  Offset = Ftell();
   char c;
   
   ReadStrLn(Buffer);		
   if(*Buffer)
   {
      if (sscanf(Buffer,"%i%c",&Var,&c) != 1)
      {
         ErrOutput << Sign 
	    << "Unable to convert '"<< Buffer <<"' into integer.\n";
         return FAILED;   
      }
      else
         return OKAY;
   }
   else
   {
      Fseek(Offset);
      return EMPTY;
   }

}


result blindBox::ReadVal(long int &Var)
{

   long int  Offset = Ftell();
   char c;
   
   ReadStrLn(Buffer);		
   if(*Buffer)
   {
      if (sscanf(Buffer,"%li%c",&Var,&c) != 1)
      {
         ErrOutput << Sign 
	    << "Unable to convert '"<< Buffer <<"' into long integer.\n";
         return FAILED;   
      }
      else
         return OKAY;
   }
   else
   {
      Fseek(Offset);
      return EMPTY;
   }

}



result blindBox::ReadVal(unsigned int &Var)
{

   long int  Offset = Ftell();
   char c;
   
   ReadStrLn(Buffer);		
   if(*Buffer)
   {
      if (sscanf(Buffer,"%u%c",&Var,&c) != 1)
      {
         ErrOutput << Sign 
	    << "Unable to convert '"<< Buffer <<"' into unsigned integer.\n";
         return FAILED;   
      }
      else
         return OKAY;
   }
   else
   {
      Fseek(Offset);
      return EMPTY;
   }

}


result blindBox::ReadVal(double &Var)
{

   long int  Offset = Ftell();
   char c;
   
   ReadStrLn(Buffer);		
   if(*Buffer)
   {
      if (sscanf(Buffer,"%lf%c",&Var,&c) != 1)
      {
         ErrOutput << Sign 
	    << "Unable to convert '"<< Buffer <<"' into double.\n";
         return FAILED;   
      }
      else
         return OKAY;
   }
   else
   {
      Fseek(Offset);
      return EMPTY;
   }

}


result blindBox::ReadVal(long double &Var)
{

   long int  Offset = Ftell();
   char c;
   
   ReadStrLn(Buffer);		
   if(*Buffer)
   {
      if (sscanf(Buffer,"%Lf%c",&Var,&c) != 1)
      {
         ErrOutput << Sign 
	    << "Unable to convert '"<< Buffer <<"' into long double.\n";
         return FAILED;   
      }
      else
         return OKAY;
   }
   else
   {
      Fseek(Offset);
      return EMPTY;
   }

}


result blindBox::ReadVal(char &Var)
{

   long int  Offset = Ftell();
   
   ReadStrLn(Buffer);	
   if(*Buffer)
   {   
      if(Buffer[1])
      {
         ErrOutput << Sign << "Unable to convert '"<< Buffer <<"' into char.\n";
         return FAILED;   
      }
      else
      {   
	 Var = *Buffer;
         return OKAY;
      }
   }
   else
   {
      Fseek(Offset);
      return EMPTY;
   }

}


result blindBox::ReadVal(char *&String)
{
   
   ReadStrLn(Buffer);

   if(*Buffer)
   {
      if(String)
      {
         // only String==NULL is allowed
         // ...it helps to avoid possible problems concerned dynamic memory
         // management
      
         WarnOutput << Sign 
	    << "Argument of the method 'ReadVal(char*&)' "
	    << "has been already defined.\n";
         Bug;
         return FAILED;
      }
      else
      {
	 if(String = strnewcpy(Buffer))

	    return OKAY;

	 else

	    return FAILED;

      }
   }
   else
      
      return EMPTY;
   
}
      


result blindBox::ReadVal(char *Id, int &Var)
{

   long int  Offset = Ftell();
   
   if(FindIdentifier(Id,SepEQU) == OKAY)
      if(ReadVal(Var) == OKAY)

	 return OKAY;
   
   Fseek(Offset);

   return EMPTY;
		     
}
   

result blindBox::ReadVal(char *Id, unsigned int &Var)
{

   long int  Offset = Ftell();
   
   if(FindIdentifier(Id,SepEQU) == OKAY)
      if(ReadVal(Var) == OKAY)

	 return OKAY;
   
   Fseek(Offset);

   return EMPTY;
		     
}


result blindBox::ReadVal(char *Id, long int &Var)
{

   long int  Offset = Ftell();
   
   if(FindIdentifier(Id,SepEQU) == OKAY)
      if(ReadVal(Var) == OKAY)

	 return OKAY;
   
   Fseek(Offset);

   return EMPTY;
		     
}
   

result blindBox::ReadVal(char *Id, double &Var)
{

   long int  Offset = Ftell();
   
   if(FindIdentifier(Id,SepEQU) == OKAY)
      if(ReadVal(Var) == OKAY)

	 return OKAY;
   
   Fseek(Offset);

   return EMPTY;
		     
}
   

result blindBox::ReadVal(char *Id, long double &Var)
{

   long int  Offset = Ftell();
   
   if(FindIdentifier(Id,SepEQU) == OKAY)
      if(ReadVal(Var) == OKAY)

	 return OKAY;
   
   Fseek(Offset);

   return EMPTY;
		     
}
   

result blindBox::ReadVal(char *Id, char &Var)
{

   long int  Offset = Ftell();
   
   if(FindIdentifier(Id,SepEQU) == OKAY)
      if(ReadVal(Var) == OKAY)

	 return OKAY;
   
   Fseek(Offset);

   return EMPTY;
		     
}


result blindBox::ReadVal(char *Id, char *&Var)
{

   long int  Offset = Ftell();
   result Res;
   
   if((Res=FindIdentifier(Id,SepEQU)) == OKAY)
      if((Res=ReadVal(Var)) == OKAY)

	 return OKAY;
   
   Fseek(Offset);

   return Res;
		     
}
   

//// ReadArray /////////////////////////////////
//

result blindBox::ReadArray(char *Id, int *&Array, int Len)
{
   
   long int  Offset = Ftell();
   int i;
   
   if(!Array)
      AllocArray(Array, Len);
   if(!Array)
      return FAILED;
	 
   if(FindIdentifier(Id,SepDDOT) == OKAY)
   {
      for(i=0;i<Len;i++)
      {
	 IgnoreSpc();
	 if(ReadVal(Array[i]) != OKAY)
	 {
	    ErrOutput << Sign 
	       << "Unable to read " << i << "-th integer from '" << Id <<"'\n";
	    return FAILED;
	 }
      }

      IgnoreSpc();
      if(CheckChar(EndMARK) == FALSE)
      {
	 WarnOutput << Sign 
	    << "End-mark check of 1-d matrix '"<< Id <<"' failed.\n";
         Fseek(Offset);
      }	 

      return OKAY;
      
   }
   
   Fseek(Offset);
   
   return EMPTY;
   
}

   
result blindBox::ReadArray(char *Id, double *&Array, int Len)
{
   
   long int  Offset = Ftell();
   int i;
   
   if(!Array)
      AllocArray(Array, Len);
   if(!Array)
      return FAILED;
	 
   if(FindIdentifier(Id,SepDDOT) == OKAY)
   {
      for(i=0;i<Len;i++)
      {
	 IgnoreSpc();
	 if(ReadVal(Array[i]) != OKAY)
	 {
	    ErrOutput << Sign 
	       << "Unable to read " << (i+1) << "-th double from '" 
	       << Id <<"'\n";
	    return FAILED;
	 }
      }

      IgnoreSpc();
      if(CheckChar(EndMARK) == FALSE)
      {
	 WarnOutput << Sign 
	    << "End-mark check of 1-d matrix '"<< Id <<"' failed.\n";
         Fseek(Offset);
      }	 

      return OKAY;
      
   }
   
   Fseek(Offset);
   
   return EMPTY;
   
}


result blindBox::ReadArray(char *Id, int **&Array, int nRow, int nCol)
{
   
   long int  Offset = Ftell();
   int row,col;
   
   if(!Array)
      AllocArray(Array, nRow, nCol);
   if(!Array)
      return FAILED;
	 
   if(FindIdentifier(Id,SepDDOT) == OKAY)
   {
      for(row=0;row<nRow;row++)
	 for(col=0;col<nCol;col++)
         {
	    IgnoreSpc();
	    if(ReadVal(Array[row][col]) != OKAY)
	    {
	       ErrOutput << Sign << "Unable to read [" << row << ',' << col
		  << "] int element from '" << Id << "' matrix.\n";
	       return FAILED;
	    }
         }

      IgnoreSpc();
      if(CheckChar(EndMARK) == FALSE)
      {
	 WarnOutput << Sign 
	    << "End-mark check of 2-d matrix '"<< Id <<"' failed.\n";
         Fseek(Offset);
      }	 

      return OKAY;
      
   }
   
   Fseek(Offset);
   
   return EMPTY;
   
}


   
result blindBox::ReadArray(char *Id, double **&Array, int nRow, int nCol)
{
   
   long int  Offset = Ftell();
   int row,col;
   
   if(!Array)
      AllocArray(Array, nRow, nCol);
   if(!Array)
      return FAILED;
	 
   if(FindIdentifier(Id,SepDDOT) == OKAY)
   {
      for(row=0;row<nRow;row++)
	 for(col=0;col<nCol;col++)
         {
	    IgnoreSpc();
	    if(ReadVal(Array[row][col]) != OKAY)
	    {
	       ErrOutput << Sign << "Unable to read [" <<row << ',' <<col
		  << "] double element from '" << Id << "' matrix.\n";
	       return FAILED;
	    }
         }

      IgnoreSpc();
      if(CheckChar(EndMARK) == FALSE)
      {
	 WarnOutput << Sign 
	    << "End-mark check of 2-d matrix '"<< Id <<"' failed.\n";
         Fseek(Offset);
      }	 

      return OKAY;
      
   }
   
   Fseek(Offset);
   
   return EMPTY;
   
}


result blindBox::ReadArray
   (char *Id, int ***&Array, int nBlock, int nRow, int nCol)
{
   
   long int  Offset = Ftell();
   int row,col,block;
   
   if(!Array)
      AllocArray(Array, nBlock, nRow, nCol);
   if(!Array)
      return FAILED;
	 
   if(FindIdentifier(Id,SepDDOT) == OKAY)
   {
      for(block=0;block<nBlock;block++)
         for(row=0;row<nRow;row++)
            for(col=0;col<nCol;col++)
            {
	       IgnoreSpc();
	       if(ReadVal(Array[block][row][col]) != OKAY)
	       {
	          ErrOutput << Sign 
		     << "Unable to read [" << block << ',' << row<< ',' << col
		     << "] int element from '" << Id << "' matrix.\n";
		  return FAILED;
	       }
            }
      IgnoreSpc();
      if(CheckChar(EndMARK) == FALSE)
      {
	 WarnOutput << Sign 
	    << "End-mark check of 3-d matrix '"<< Id <<"' failed.\n";
         Fseek(Offset);
      }	 

      return OKAY;
      
   }
   
   Fseek(Offset);
   
   return EMPTY;
   
}


result blindBox::ReadArray
   (char *Id, double ***&Array, int nBlock, int nRow, int nCol)
{
   
   long int  Offset = Ftell();
   int row,col,block;
   
   if(!Array)
      AllocArray(Array, nBlock, nRow, nCol);
   if(!Array)
      return FAILED;
	 
   if(FindIdentifier(Id,SepDDOT) == OKAY)
   {
      for(block=0;block<nBlock;block++)
         for(row=0;row<nRow;row++)
            for(col=0;col<nCol;col++)
            {
	       IgnoreSpc();
	       if(ReadVal(Array[block][row][col]) != OKAY)
	       {
	          ErrOutput << Sign 
		     << "Unable to read [" << block << ',' << row << ',' << col
		     << "] double element from '" << Id << "' matrix.\n";
		  return FAILED;
	       }
            }
      IgnoreSpc();
      if(CheckChar(EndMARK) == FALSE)
      {
	 WarnOutput << Sign 
	    << "End-mark check of 3-d matrix '"<< Id <<"' failed.\n";
         Fseek(Offset);
      }	 

      return OKAY;
      
   }
   
   Fseek(Offset);
   
   return EMPTY;
   
}


//// blindBox::Putc /////////////////////////////////
//

void blindBox::Putc(char c)
{

   OpenW();
   if(putc(c,File) == EOF)
   {
      ErrNOutput(errno) << Sign 
	 << "Unable to get the position in '" << FileName << "'\n";
      exit(errno);
    }
   else
      return;

}


//// blindBox::WriteStr /////////////////////////////////
//

result blindBox::WriteStr(char *Str)
{

   OpenW();
   if(Str)
      if(strlen(Str) != fprintf(File,"%s",Str))
	 {
	    ErrOutput << Sign
	       << "The string '" << Str <<"' not written into '"
	       << FileName <<"'\n";
	    return FAILED;
	 }

   return OKAY;

}


//// blindBox::WriteSection /////////////////////////////////
//

result blindBox::WriteSection()
{
   
   return WriteSection(Signature);
   
}


result blindBox::WriteSection(char *Name)
{

   WriteLn();
   Putc(LeftSEC);
   Putc(' ');
   WriteStr(Name);
   Putc(' ');
   Putc(RightSEC);
   WriteLn();
   
   return OKAY;
   
}


//// blindBox::WriteIdent /////////////////////////////////
//

result blindBox::WriteIdent(char *Id, char Separ)
{
   OpenW();
   int res = WriteStr(Id);
   Putc(' ');
   Putc(Separ);
   Putc(' ');
   
   return res;
   
}


////  blindBox::WriteVal /////////////////////////////////
//

result blindBox::WriteVal(int Val, char *Format)
{

   if(Format)
   {
      OpenW();
      if(fprintf(File,Format,Val) > 0)
	 return OKAY;
      else
      {
	 ErrOutput << Sign 
	    << "Writing int value '" << Val << "' into '"
	    << FileName <<"' failed.\n";
	 return FAILED;
      }   
   }
   else 
      return EMPTY;
   
}


result blindBox::WriteVal(unsigned int Val, char *Format)
{

   if(Format)
   {
      OpenW();
      if(fprintf(File,Format,Val) > 0)
	 return OKAY;
      else
      {
	 ErrOutput << Sign 
	    << "Writing unsigned int value '" << Val << "' into '"
	    << FileName <<"' failed.\n";
	 return FAILED;
      }   
   }
   else 
      return EMPTY;
   
}


result blindBox::WriteVal(long int Val, char *Format)
{

   if(Format)
   {
      OpenW();
      if(fprintf(File,Format,Val) > 0)
	 return OKAY;
      else
      {
	 ErrOutput << Sign << "Writing long int value '" << Val <<"' into '"\
	    << FileName <<"' failed.\n";
	 return FAILED;
      }   
   }
   else 
      return EMPTY;
   
}


result blindBox::WriteVal(double Val, char *Format)
{

   if(Format)
   {
      OpenW();
      if(fprintf(File,Format,Val) > 0)
	 return OKAY;
      else
      {
	 ErrOutput << Sign << "Writing double value '"<< Val <<"' into '"\
	    << FileName <<"' failed.\n";
	 return FAILED;
      }   
   }
   else 
      return EMPTY;
   
}


result blindBox::WriteVal(long double Val, char *Format)
{

   if(Format)
   {
      OpenW();
      if(fprintf(File,Format,Val) > 0)
	 return OKAY;
      else
      {
	 ErrOutput << Sign << "Writing long double value into '"\
	    << FileName <<"' failed.\n";
	 return FAILED;
      }   
   }
   else 
      return EMPTY;
   
}


result blindBox::WriteVal(char *Id, int Val, char *Format)
{

   int res;
   
   WriteIdent(Id,SepEQU);
   if(Format)
      res = WriteVal(Val,Format);
   else
      res = WriteVal(Val);
   
   return res;
   
}


result blindBox::WriteVal(char *Id, unsigned int Val, char *Format)
{

   int res;
   
   WriteIdent(Id,SepEQU);
   if(Format)
      res = WriteVal(Val,Format);
   else
      res = WriteVal(Val);
   
   return res;
   
}


result blindBox::WriteVal(char *Id, long int Val, char *Format)
{

   int res;
   
   WriteIdent(Id,SepEQU);
   if(Format)
      res = WriteVal(Val,Format);
   else
      res = WriteVal(Val);
   
   return res;
   
}


result blindBox::WriteVal(char *Id, double Val, char *Format)
{

   int res;
   
   WriteIdent(Id,SepEQU);
   if(Format)
      res = WriteVal(Val,Format);
   else
      res = WriteVal(Val);
   
   return res;
   
}


result blindBox::WriteVal(char *Id, long double Val, char *Format)
{

   int res;
   
   WriteIdent(Id,SepEQU);
   if(Format)
      res = WriteVal(Val,Format);
   else
      res = WriteVal(Val);
   
   return res;
   
}


result blindBox::WriteVal(char *Id, char *Var)
{
   
   WriteIdent(Id,SepEQU);

   return WriteStr(Var);
   
}


////  blindBox::WriteArray /////////////////////////////////
//

result blindBox::WriteArray(char *Id, int *Array, int Len, char *Format)
{

   int row;

   WriteIdent(Id,SepDDOT);
   WriteLn();

   for(row=0;row<Len;row++)
   {
      Putc('\t');
      if(Format)
	 WriteVal(Array[row],Format);
      else
	 WriteVal(Array[row]);
      WriteLn();
   }
      
   Putc(EndMARK);
   WriteLn();
   
   return OKAY;
   
}


result blindBox::WriteArray
   (char *Id, int **Array, int nRow, int nCol, char *Format)
{

   int col,row;

   WriteIdent(Id,SepDDOT);
   WriteLn();

   for(row=0;row<nRow;row++)
   {
      for(col=0;col<nCol;col++)
      {  
	 WriteTab();
         if(Format)
	    WriteVal(Array[row][col],Format);
         else
	    WriteVal(Array[row][col]);
      }
      WriteLn();
   }
      
   Putc(EndMARK);
   WriteLn();
   
   return OKAY;
   
}


result blindBox::WriteArray
   (char *Id, int ***Array, int nBlock, int nRow, int nCol, char *Format)
{

   int col,row,block;

   WriteIdent(Id,SepDDOT);

   for(block=0;block<nBlock;block++)
   {
      WriteLn();
      for(row=0;row<nRow;row++)
      {
         for(col=0;col<nCol;col++)
         {  
	    WriteTab();
            if(Format)
	       WriteVal(Array[block][row][col],Format);
            else
	       WriteVal(Array[block][row][col]);
         }
         WriteLn();
      }
   }
      
   Putc(EndMARK);
   WriteLn();
   
   return OKAY;
   
}


result blindBox::WriteArray(char *Id, double *Array, int Len, char *Format)
{

   int row;

   WriteIdent(Id,SepDDOT);
   WriteLn();

   for(row=0;row<Len;row++)
   {
      Putc('\t');
      if(Format)
	 WriteVal(Array[row],Format);
      else
	 WriteVal(Array[row]);
      WriteLn();
   }
      
   Putc(EndMARK);
   WriteLn();
   
   return OKAY;
   
}


result blindBox::WriteArray
   (char *Id, double **Array, int nRow, int nCol, char *Format)
{

   int col,row;

   WriteIdent(Id,SepDDOT);
   WriteLn();

   for(row=0;row<nRow;row++)
   {
      for(col=0;col<nCol;col++)
      {  
	 WriteTab();
         if(Format)
	    WriteVal(Array[row][col],Format);
         else
	    WriteVal(Array[row][col]);
      }
      WriteLn();
   }
      
   Putc(EndMARK);
   WriteLn();
   
   return OKAY;
   
}


result blindBox::WriteArray
   (char *Id, double ***Array, int nBlock, int nRow, int nCol, char *Format)
{

   int col,row,block;

   WriteIdent(Id,SepDDOT);

   for(block=0;block<nBlock;block++)
   {
      WriteLn();
      for(row=0;row<nRow;row++)
      {
         for(col=0;col<nCol;col++)
         {  
	    WriteTab();
            if(Format)
	       WriteVal(Array[block][row][col],Format);
            else
	       WriteVal(Array[block][row][col]);
         }
         WriteLn();
      }
   }
      
   Putc(EndMARK);
   WriteLn();
   
   return OKAY;
   
}



////  blindBox::strnewcpy /////////////////////////////////
//

char *blindBox::strnewcpy(const char *s)
{

   char *ns;
   
   if(ns=new char[strlen(s)+1])
      strcpy(ns,s);
   else
      ErrOutput<< Sign << MallocErr;
   
   return ns;

}



////  blindBox::AllocArray /////////////////////////////////
//

void blindBox::AllocArray(int *&A, unsigned N)
{

   if(A)
   {
      ErrOutput << Sign << "Array has been already defined.\n";
      Bug;
      exitus;
   }
   if(!(A=new int[N]))
      ErrOutput << Sign << MallocErr;
   
   return;
   
}



void blindBox::AllocArray(float *&A, unsigned N)
{

   if(A)
   {
      ErrOutput << Sign << "Array has been already defined.\n";
      Bug;
      exitus;
   }
   if(!(A=new float[N]))
      ErrOutput << Sign << MallocErr;
   
   return;
   
}



void blindBox::AllocArray(double *&A, unsigned N)
{

   if(A)
   {
      ErrOutput << Sign << "Array has been already defined.\n";
      Bug;
      exitus;
   }
   if(!(A=new double[N]))
      ErrOutput << Sign << MallocErr;
   
   return;
   
}



void blindBox::AllocArray(int **&A, unsigned Nrow, unsigned Ncol)
{

   int r,rr;
   
   if(A)
   {
      ErrOutput << Sign << "Array has been already defined.\n";
      Bug;
      exitus;
   }
   
   if(A=new int*[Nrow])
      for(r=0; r<Nrow; r++)
	 if(!(A[r] = new int[Ncol]))
	 {
	    // deallocate
	    for(rr=0; rr<r;rr++)
	       delete [] (A[rr]);
	    delete [] A;
	    A = NULL;
	    break;
	 }
   
   if(!A)
      ErrOutput << Sign << MallocErr;
   
   return;
   
}



void blindBox::AllocArray(float **&A, unsigned Nrow, unsigned Ncol)
{

   int r,rr;
   
   if(A)
   {
      ErrOutput << Sign << "Array has been already defined.\n";
      Bug;
      exitus;
   }
   
   if(A=new float*[Nrow])
      for(r=0; r<Nrow; r++)
	 if(!(A[r] = new float[Ncol]))
	 {
	    // deallocate
	    for(rr=0; rr<r;rr++)
	       delete [] (A[rr]);
	    delete [] A;
	    A = NULL;
	    break;
	 }
   
   if(!A)
      ErrOutput << Sign << MallocErr;
   
   return;
   
}



void blindBox::AllocArray(double **&A, unsigned Nrow, unsigned Ncol)
{

   int r,rr;
   
   if(A)
   {
      ErrOutput << Sign << "Array has been already defined.\n";
      Bug;
      exitus;
   }
   
   if(A=new double*[Nrow])
      for(r=0; r<Nrow; r++)
	 if(!(A[r] = new double[Ncol]))
	 {
	    // deallocate
	    for(rr=0; rr<r;rr++)
	       delete [] (A[rr]);
	    delete [] A;
	    A = NULL;
	    break;
	 }
   
   if(!A)
      ErrOutput << Sign << MallocErr;
   
   return;
   
}



void blindBox::AllocArray(int ***&A, unsigned Nblock, unsigned Nrow, unsigned Ncol)
{

   int b,bb;
   
   if(A)
   {
      ErrOutput << Sign << "Array has been already defined.\n";
      Bug;
      exitus;
   }
   
   if(A=new int**[Nblock])
      for(b=0; b<Nblock; b++)
      {	 
	 AllocArray(A[b], Nrow, Ncol);
	 if(!A[b])
	 {
	    // deallocate
	    for(bb=0; bb<b;bb++)
	       DeallocArray(A[bb],Nrow);
	    delete [] A;
	    A = NULL;
	    break;
	 }
      }
   
   if(!A)
      ErrOutput << Sign << MallocErr;
   
   return;
   
}



void blindBox::AllocArray(float ***&A, unsigned Nblock, unsigned Nrow, unsigned Ncol)
{

   int b,bb;
   
   if(A)
   {
      ErrOutput << Sign << "Array has been already defined.\n";
      Bug;
      exitus;
   }
   
   if(A=new float**[Nblock])
      for(b=0; b<Nblock; b++)
      {
	 AllocArray(A[b], Nrow, Ncol);
	 if(!A[b])
	 {
	    // deallocate
	    for(bb=0; bb<b;bb++)
	       DeallocArray(A[bb],Nrow);
	    delete [] A;
	    A = NULL;
	    break;
	 }
      }
   
   if(!A)
      ErrOutput << Sign << MallocErr;
   
   return;
   
}



void blindBox::AllocArray(double ***&A, unsigned Nblock, unsigned Nrow, unsigned Ncol)
{

   int b,bb;
   
   if(A)
   {
      ErrOutput << Sign << "Array has been already defined.\n";
      Bug;
      exitus;
   }
   
   if(A=new double**[Nblock])
      for(b=0; b<Nblock; b++)
      {
	 AllocArray(A[b], Nrow, Ncol);
	 if(!A[b])
	 {
	    // deallocate
	    for(bb=0; bb<b;bb++)
	       DeallocArray(A[bb],Nrow);
	    delete [] A;
	    A = NULL;
	    break;
	 }
      }
   
   if(!A)
      ErrOutput << Sign << MallocErr;
   
   return;
   
}



////  blindBox::DeallocArray /////////////////////////////////
//

void blindBox::DeallocArray(int *A)
{

   if(A)
      delete [] A;
   
   return;

}   
 

void blindBox::DeallocArray(float *A)
{

   if(A)
      delete [] A;
   
   return;

}   
 

   
void blindBox::DeallocArray(double *A)
{

   if(A)
      delete [] A;
   
   return;

}   
 

      
void blindBox::DeallocArray(int **A, unsigned Nrow)
{

   int r;
   
   if(A)
   {
      for(r=0;r<Nrow;r++)
	 delete [] (A[r]);
      delete [] A;
      A = NULL;
   }
   
   return;

}   
 

      
void blindBox::DeallocArray(float **A, unsigned Nrow)
{

   int r;
   
   if(A)
   {
      for(r=0;r<Nrow;r++)
	 delete [] (A[r]);
      delete [] A;
      A = NULL;
   }
   
   return;

}   
 

      
void blindBox::DeallocArray(double **A, unsigned Nrow)
{

   int r;
   
   if(A)
   {
      for(r=0;r<Nrow;r++)
	 delete [] (A[r]);
      delete [] A;
      A = NULL;
   }
   
   return;

}   
 

      
void blindBox::DeallocArray(int ***A, unsigned Nblock, unsigned Nrow)
{

   int b;
   
   if(A)
   {
      for(b=0;b<Nblock;b++)
	 DeallocArray(A[b], Nrow);
      delete [] A;
      A = NULL;
   }
   
   return;

}   
 

      
void blindBox::DeallocArray(float ***A, unsigned Nblock, unsigned Nrow)
{

   int b;
   
   if(A)
   {
      for(b=0;b<Nblock;b++)
	 DeallocArray(A[b], Nrow);
      delete [] A;
      A = NULL;
   }
   
   return;

}   
 

      
void blindBox::DeallocArray(double ***A, unsigned Nblock, unsigned Nrow)
{

   int b;
   
   if(A)
   {
      for(b=0;b<Nblock;b++)
	 DeallocArray(A[b], Nrow);
      delete [] A;
      A = NULL;
   }
   
   return;

}   
 

      
////  blindBox::Backup /////////////////////////////////
//

result blindBox::Backup(char *BckFName)
{

   int c;
   FILE	*BckFile;
   
   
   if(!BckFName)
        // file name is empty
      return EMPTY;
	 
	 
   if (!(BckFile=fopen(BckFName,"w")))
   {
      WarnNOutput(errno) << Sign << FopenErr(BckFName);
      return FAILED;
   } 

   Close();
   OpenR();

   while((c=Getc()) != EOF)
      putc(c,BckFile);

   if (fclose(BckFile))
   {
      WarnNOutput(errno) << Sign << FcloseErr(BckFName);
      return FAILED;
   }

   
   return OKAY;
   
}
