//////////////////////////////////////////////////////////////////////////////
//
// File:	blind.h
// Directory:	/users/hartmann/rapid/src
//
//////////////////////////////////////////////////////////////////////////////

#ifndef BLIND_H
#define BLIND_H



#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "defs.h"
#include "errtool.h"

//key characters
extern const char	SepEQU;
extern const char	SepDDOT;
extern const char	EndMARK;
extern const char	LeftSEC;
extern const char	RightSEC;

//char buffer length
#define	BufferSIZE 512



//////////////////////////////////////////////////////////////////////////////
//
// CLASS: blindBox
//
//////////////////////////////////////////////////////////////////////////////

class blindBox
{
   
   friend class kernel;

   protected:
      static char		Buffer[BufferSIZE];
      char			*Signature;
      static char 		*FileName;

   private:
      static FILE		*File;
      static int		FStatus;
      static unsigned int	InitFlag;
      static long int		SectionOffset;
      
   public:   
   //protected:
      blindBox(const char*, char*);
      blindBox(const char*);

      char Getc();
      void Ungetc();
      result Backup(char*);	// copies *File into file specifid by name

      boolean CheckChar(char);
      boolean CheckStr(char*);
      void IgnoreSpc();
      void IgnoreSpcLn();
      void IgnoreSet(const char*);
      void GotoNextLn();

      result FindIdentifier(char *, char);

      result ReadVal();
      result ReadVal(int&);
      result ReadVal(unsigned int&);
      result ReadVal(long int&);
      result ReadVal(double&);
      result ReadVal(long double&);
      result ReadVal(char&);
      result ReadVal(char*&);

      result ReadVal(char*, int&);
      result ReadVal(char*, unsigned int&);
      result ReadVal(char*, long int&);
      result ReadVal(char*, double&);
      result ReadVal(char*, long double&);
      result ReadVal(char*, char&);
      result ReadVal(char*, char*&);
  
      result ReadArray(char*, int*&, int); 
      result ReadArray(char*, double*&, int);
      result ReadArray(char*, int **&, int, int);
      result ReadArray(char*, double **&, int, int);
      result ReadArray(char*, int ***&, int, int, int);
      result ReadArray(char*, double ***&, int, int, int);
      
      void Putc(char);
      void WriteLn() { Putc('\n'); };
      void WriteTab() { Putc('\t'); };

      result WriteStr(char*);
      result WriteSection();
      result WriteSection(char*);
      result WriteIdent(char*,char);
      
      result WriteVal(int, char *Format="%i");
      result WriteVal(unsigned int, char *Format="%u");		      
      result WriteVal(long int, char *Format="%li");		      
      result WriteVal(double, char *Format="%g");		      
      result WriteVal(long double, char *Format="%Lf");		      
      result WriteVal(char *Var) { return WriteStr(Var); };
      
      result WriteVal(char*, int, char *Format=NULL);
      result WriteVal(char*, unsigned int, char *Format=NULL);
      result WriteVal(char*, long int, char *Format=NULL);
      result WriteVal(char*, double, char *Format=NULL);
      result WriteVal(char*, long double, char *Format=NULL);
      result WriteVal(char*, char*);
      
      result WriteArray(char*, int*, int, char *Format=NULL);
      result WriteArray(char*, int**, int, int, char *Format=NULL);
      result WriteArray(char*, int***, int, int, int, char *Format=NULL);
      result WriteArray(char*, double*, int, char *Format=NULL);
      result WriteArray(char*, double**, int, int, char *Format=NULL);
      result WriteArray(char*, double***, int, int, int, char *Format=NULL);
      
      char *strnewcpy(const char*);

      void AllocArray(int*&, unsigned N);
      void AllocArray(float*&, unsigned N);
      void AllocArray(double*&, unsigned N);

      void AllocArray(int**&, unsigned Nrow, unsigned Ncol);
      void AllocArray(float**&, unsigned Nrow, unsigned Ncol);
      void AllocArray(double**&, unsigned Nrow, unsigned Ncol);

      void AllocArray(int***&,unsigned Nblock,unsigned Nrow,unsigned Ncol);
      void AllocArray(float***&,unsigned Nblock,unsigned Nrow,unsigned Ncol);
      void AllocArray(double***&,unsigned Nblock,unsigned Nrow,unsigned Ncol);

      void DeallocArray(int*);
      void DeallocArray(float*);
      void DeallocArray(double*);

      void DeallocArray(int**, unsigned Nrow);
      void DeallocArray(float**, unsigned Nrow);
      void DeallocArray(double**, unsigned Nrow);

      void DeallocArray(int***,unsigned Nblock,unsigned Nrow);
      void DeallocArray(float***,unsigned Nblock,unsigned Nrow);
      void DeallocArray(double***,unsigned Nblock,unsigned Nrow);

   //private:    
      result FindSection(const char *SectionId = NULL);
      void ReadStrLn(char*);

      void Fseek(long int);
      long int Ftell();
      
      void Close();
      void OpenR();
      void OpenW();

};
      


#endif
