//////////////////////////////////////////////////////////////////////////////
//
// File:        mlpkrnl.C
// Directory:   /users/pbienst/mlp   
//
//////////////////////////////////////////////////////////////////////////////


#include "mlpkrnl.h"


//system signals
const sysig NOsig  = 0;
const sysig Fsig   = 1;
const sysig Bsig   = 2;
const sysig Errsig = 4;

const outp desired   = 0;
const outp computed  = 1;
const outp error     = 2;
const outp distance  = 3;
const outp tdistance = 4;
const outp input     = 5;

//key words
const char *Kernel        = "mlp";
const char *TopologyKey   = "Topology";
const char *BckFileKey    = "BackupFile";
const char *SaveCycKey	  = "SavePeriod";
const char *SaveEnKey	  = "SaveEnable";
const char *TrueKey       = "true";
const char *FalseKey      = "false";
const char *SeedKey       = "SeedRand";
const char *IndexOnlyKey  = "ShowOnlyIndex";
const char *NBestKey      = "N-best";
const char *OutputKey     = "Output";
const char *OutputKeys[6] = { "desired", "computed", "error",
                              "distance","tdistance", "input" };
   
//random generator's seedvalue
const long int SeedInitVal = 0l;


//// kernel::kernel /////////////////////////////////
//

kernel::kernel(const char *SignStr, char *FName) 
   : blindBox(SignStr,FName) {};
   


//// kernel::init_kernel ////////////////////////////////////
//
   
// Reads in information in the kernel section of the specification file.
   

void kernel::init_kernel()
{

   //
   //  find kernel section
   //
   
   
   if(FindSection(Kernel) != OKAY)
   {
      ErrOutput << Sign 
	 << "Unable to find section '" << Kernel << "'\n";
      exitus;
   }
   
      
   //
   //  read topology
   //
   
   
   if (FindIdentifier((char*)TopologyKey,SepEQU) != OKAY)
   {
      ErrOutput << Sign
	 << "Unable to find identifier '" << TopologyKey << "'\n";
      exitus;
   }
   
   layers=0;
   while (ReadVal(neurons[layers]) == OKAY)
   {
 	if (layers==MAXLAYERS)
	{
	 	ErrOutput << Sign << "Too many layers.\n";
		ErrOutput << Sign << "Increase MAXLAYERS and recompile.\n";
		exitus;
	}
	layers++;
   }
   
   if (layers < 2)
   {
      ErrOutput << Sign << "Minimum two layers.\n";
      exitus;
   }
   
   
   //
   //   read other kernel variables
   //
   
   
   if(ReadVal((char*)BckFileKey,BackupFName) != OKAY)
      BackupFName = NULL;
   
   if(ReadVal((char*)SaveCycKey, SaveCycle) != OKAY)
      SaveCycle = 0;
   
   
   SaveEnable = FALSE;
   if(FindIdentifier((char*)SaveEnKey,SepEQU) == OKAY)
   {
      IgnoreSpcLn();
      if(CheckStr((char*)TrueKey) == TRUE)
      {
	 SaveEnable = TRUE;
	 if (SaveCycle==0)
	    WarnOutput << Sign << SaveCycKey << " was set to zero\n";
      }
      else if(CheckStr((char*)FalseKey) != TRUE)
	 WarnOutput << Sign << "Guessing " 
	    << SaveEnKey << SepEQU << FalseKey <<'\n';
   }

   SaveCntr=0;
   
     
   if(ReadVal((char*)SeedKey,SeedVal) != OKAY)
      SeedVal = SeedInitVal;
   seedrand(SeedVal);
   
   IndexOnly = FALSE;
   if(FindIdentifier((char*)IndexOnlyKey,SepEQU) == OKAY)
   {
      IgnoreSpcLn();
      if(CheckStr((char*)TrueKey) == TRUE)
	 IndexOnly = TRUE;

      else if(CheckStr((char*)FalseKey) != TRUE)
	 WarnOutput << Sign << "Guessing " 
	    << IndexOnlyKey << SepEQU << TrueKey <<'\n';
   }
   
   if(ReadVal((char*)NBestKey,Nbest) != OKAY)
      Nbest = 1;
   
   
   //
   // read output field (used in train and test)
   //
   
   
   for (outp j=desired; j <= input; j++) output[j]=-1; // undefined
   
   long int StartOffset, FoundOffset;
   
   if(FindIdentifier((char*)OutputKey,SepEQU) == OKAY)
   {
      
      for (int i=0; i<6; i++)
      {
	 StartOffset = Ftell();
	 int Found = FALSE;
	 for (int j=desired; j <= input; j++ )
	 {
	    IgnoreSpcLn();
	    if ( CheckStr((char*)OutputKeys[j]) == TRUE )
	       if (Found == FALSE) // nothing found before
	       {
		  output[i] = j;
		  Found = TRUE;
		  FoundOffset = Ftell();
		  Fseek(StartOffset);
	       }
	       else if (Ftell() < FoundOffset) // earlier in list ?
	       {
		  output[i] = j;
		  FoundOffset = Ftell();
		  Fseek(StartOffset);
	       }
	 }
	 
	 if (Found == TRUE) Fseek(FoundOffset);
      }
      
   }
  
}
   

//// kernel::config_general //////////////////////////
//

// Sets up the boxes al the subprograms have in
// common, i.e. input and actual neural network.

void kernel::config_general()
{
   boxes=0;

   //
   //  vector in
   //
   
   create_box((char*)Vectorin,"in", 0, 1, 1, 0);
   // the correct size will be determined in the procedure
 
   InB = boxes-1;
   StartNeural = boxes; 
   
   Boxlist[InB]->Yout->Port[0] = 0;
   Boxlist[InB]->Yout->Size[0] = neurons[0];
   if (!(Boxlist[InB]->Yout->Data[0] = new signal[neurons[0]] ))
   {
      ErrOutput << MallocErr;
      exitus;
   }
   
   Boxlist[InB]->Yin->Port[0] = 0;
   Boxlist[InB]->Yin->Size[0] = neurons[0];
   if (!(Boxlist[InB]->Yin->Data[0] = new signal[neurons[0]] ))
   {
      ErrOutput << MallocErr;
      exitus;
   }
   
   // "To" connections on the Y side have to be specified later
   
   // the box itself will be initialised later, together with the other
   // boxes that read their information from the kernel section
   
   
   //
   //   actual neural network
   //
   
   
   int CurrentLayer=0, size =0;
   char	Buffer[BufferSIZE];
   
   while (ReadNextSection(Buffer) == OKAY)
   {
      
     //-- either create linear box
      
     if ( (!(strcmp(Buffer,Synapses)))    ||
	  (!(strcmp(Buffer,SynapsesTD)))  ||
	  (!(strcmp(Buffer,SynapsesTD2)))
	)
     {
	create_box(Buffer, Buffer, 1, 1, 1, 1);
	
	Boxlist[boxes-1]->Xin ->Size[0] = neurons[CurrentLayer];
	Boxlist[boxes-1]->Xout->Size[0] = neurons[CurrentLayer++];
	
	if (CurrentLayer == layers)
	{
	   ErrOutput << Sign << "box sequence and topology incompatible.\n";
	   exitus;
	}
	
	Boxlist[boxes-1]->Yin ->Size[0] = neurons[CurrentLayer];
	Boxlist[boxes-1]->Yout->Size[0] = neurons[CurrentLayer];
	
	size=neurons[CurrentLayer]; // for later use

     }
     
     //-- or create non linear box
     
     else if ( (!(strcmp(Buffer,Logistic)))    ||
               (!(strcmp(Buffer,BipSigmoid)))  ||
	       (!(strcmp(Buffer,SoftMax)))
             )
     {
	create_box(Buffer, Buffer, 1, 1, 1, 1);
	
        Boxlist[boxes-1]->Xin ->Size[0] = neurons[CurrentLayer];
	Boxlist[boxes-1]->Xout->Size[0] = neurons[CurrentLayer];
	Boxlist[boxes-1]->Yin ->Size[0] = neurons[CurrentLayer];
	Boxlist[boxes-1]->Yout->Size[0] = neurons[CurrentLayer];
	
	size=neurons[CurrentLayer]; // for later use
	
     }
     
     else
     {
	ErrOutput << Sign << "Unknown box type.\n";
	exitus;
     }
     
     //-- these connections are the same for linear and non linear boxes
    
     Boxlist[boxes-1]->Xin->From[0] = Boxlist[boxes-2];
     Boxlist[boxes-1]->Xin->Port[0] = 0;
     Boxlist[boxes-1]->Xin->Data[0] = Boxlist[boxes-2]->Yout->Data[0];
     
     Boxlist[boxes-1]->Xout->To[0]   = Boxlist[boxes-2];
     Boxlist[boxes-1]->Xout->Port[0] = 0;
     Boxlist[boxes-1]->Xout->Data[0] = Boxlist[boxes-2]->Yin->Data[0];
     
     Boxlist[boxes-1]->Yout->Port[0] = 0;
     if (!(Boxlist[boxes-1]->Yout->Data[0] = new signal[size]))
     {
	ErrOutput << MallocErr;
	exitus;
     }
     
     Boxlist[boxes-1]->Yin->Port[0] = 0;
     if (!(Boxlist[boxes-1]->Yin->Data[0] = new signal[size]))
     {
	ErrOutput << MallocErr;
	exitus;
     }
     
     // now we have sufficient data to initialise the box
     
     if (Boxlist[boxes-1]->init() != OKAY)
     {
	ErrOutput << Sign << "Initialisation of block " 
	   << Boxlist[boxes-1]->Signature << " failed.\n";
	exitus;
     }
     
     // actual connections on the Y side will have to be made later
     
   }
   
   if (CurrentLayer != layers-1)
   {
      ErrOutput << Sign << "box sequence and topology incompatible.\n";
      exitus;
   }
   
   EndNeural = boxes-1;
   

   //
   // connecting all boxes on the Y side
   //  
   
      
   for (int i=InB; i<=EndNeural; i++)
   {
      Boxlist[i]->Yout->To[0]  = Boxlist[i+1];
      Boxlist[i]->Yin->From[0] = Boxlist[i+1];
   }
   
   
}

//// kernel::config_io ///////////////////////////////
//
   
// Sets up the boxes specific to the io subprogram,
// in this case just a vector out block.

void kernel::config_io() 
   
{
   
   create_box((char*)Vectorout, "out", 1, 0, 0, 1); OutB = boxes-1;
   
   Boxlist[EndNeural]->Yout->To[0] = Boxlist[OutB];
  
   Boxlist[OutB]->Xin->From[0] = Boxlist[EndNeural];
   Boxlist[OutB]->Xin->Port[0] = 0;
   Boxlist[OutB]->Xin->Size[0] = neurons[layers-1];
   Boxlist[OutB]->Xin->Data[0] = Boxlist[EndNeural]->Yout->Data[0];
   
   Boxlist[OutB]->Xout->To[0]   = Boxlist[InB];
   Boxlist[OutB]->Xout->Port[0] = 0;
   Boxlist[OutB]->Xout->Size[0] = 0;
   

  
}
   
//// kernel::config_train() /////////////////////////////////////
//

// Sets up the boxes specific to the train subprogram,
// in this case a teacher and a vector out block.
// Some additional connections from vector in are made as well 



void kernel::config_train()
{
  
   create_box((char*)Teacher,"teacher", 3, 6, 1, 1 );
   create_box((char*)Vectorout, "out", 6, 0, 0, 1); OutB = boxes-1;
 
   //   
   // connecting boxes
   //
   
     
   //-- vector in
   
   Boxlist[InB]->Yout->To[1]   = Boxlist[EndNeural+1]; // teacher
   Boxlist[InB]->Yout->Port[1] = 0;
   Boxlist[InB]->Yout->Size[1] = neurons[layers-1];
   if (!(Boxlist[InB]->Yout->Data[1] = new signal[neurons[layers-1]] ))
   {
      ErrOutput << MallocErr;
      exitus;
   }
   
   //-- end of actual NN
   
   Boxlist[EndNeural]->Yout->To[0] = Boxlist[EndNeural+1];
      
   //-- teacher
   
   //---- Xin
   Boxlist[EndNeural+1]->Xin->From[0] = Boxlist[InB];  // desired values
   Boxlist[EndNeural+1]->Xin->Port[0] = 1;
   Boxlist[EndNeural+1]->Xin->Size[0] = neurons[layers-1];
   Boxlist[EndNeural+1]->Xin->Data[0] = Boxlist[InB]->Yout->Data[1];
   
   Boxlist[EndNeural+1]->Xin->From[1] = Boxlist[EndNeural];  // actual values
   Boxlist[EndNeural+1]->Xin->Port[1] = 0;
   Boxlist[EndNeural+1]->Xin->Size[1] = neurons[layers-1];
   Boxlist[EndNeural+1]->Xin->Data[1] = Boxlist[EndNeural]->Yout->Data[0];
   
   Boxlist[EndNeural+1]->Xin->From[2] = Boxlist[InB];  // input
   Boxlist[EndNeural+1]->Xin->Port[2] = 0;
   Boxlist[EndNeural+1]->Xin->Size[2] = neurons[0];
   Boxlist[EndNeural+1]->Xin->Data[2] = Boxlist[InB]->Yout->Data[0];
   
   //---- Yout
   int i;
   for (i=desired; i<=input; i++)
   {
      
      Boxlist[EndNeural+1]->Yout->To[i]   = Boxlist[OutB]; 
      Boxlist[EndNeural+1]->Yout->Port[i] = i;
       
      int size;
      
      switch (i)
      {
	 case desired:
	 case computed:
	 case error:     size = neurons[layers-1];
	                 break;
	 case distance:
	 case tdistance: size = 1;
	                 break;
	 case input:     size = neurons[0];
      }
	       
      Boxlist[EndNeural+1]->Yout->Size[i] = size;
      
      if (!(Boxlist[EndNeural+1]->Yout->Data[i] = new signal[size] ))
      {
         ErrOutput << MallocErr;
         exitus;
      }

   }
   
   //---- Yin
 
   Boxlist[EndNeural+1]->Yin->From[0] = Boxlist[OutB];
   Boxlist[EndNeural+1]->Yin->Port[0] = 0;
   Boxlist[EndNeural+1]->Yin->Size[0] = 0;
  
   //---- Xout
   Boxlist[EndNeural+1]->Xout->To[0]   = Boxlist[EndNeural]; 
   Boxlist[EndNeural+1]->Xout->Port[0] = 0;
   Boxlist[EndNeural+1]->Xout->Size[0] = neurons[layers-1];
   Boxlist[EndNeural+1]->Xout->Data[0] = Boxlist[EndNeural]->Yin->Data[0];
   
     
   //-- vector out
  
   Boxlist[OutB]->Xin->N = 0;
   
   for (i=desired; i<=input; i++)
   {
      
     Boxlist[OutB]->Xin->From[i] = Boxlist[EndNeural+1];
     Boxlist[OutB]->Xin->Port[i] = output[i];
         
     int size;
     switch (output[i])
     {
	case desired:
	case computed:
	case error:     size=neurons[layers-1];
	                Boxlist[OutB]->Xin->N++;
	                break;
	case distance:
	case tdistance: size=1;
	                Boxlist[OutB]->Xin->N++;
	                break;
	case input:     size = neurons[0];
	                Boxlist[OutB]->Xin->N++;
	                break;
	default:        size = 0;
     }
	   
   
     Boxlist[OutB]->Xin->Size[i] = size;
     Boxlist[OutB]->Xin->Data[i] = Boxlist[EndNeural+1]->Yout->Data[output[i]];
      
   } 
      
      
   Boxlist[OutB]->Xout->To[0]   = Boxlist[EndNeural+1];
   Boxlist[OutB]->Xout->Port[0] = 0;
   Boxlist[OutB]->Xout->Size[0] = 0;
   
   create_box((char*)ConfusionDummy, "confusiondummy", 0, 0, 0, 0);
   // the purpose of creating this box is to allow the confusion
   // specific data in the kernel section to be read in, so that
   // it can be retained while saving
   

}
 
//// kernel::config_test ///////////////////////////
//

// Sets up the boxes specific to the test subprogram,
// in this case a tester and a vector out
// Additional needed connections are made well 


void kernel::config_test()
{
  
   create_box((char*)Tester,"tester", 3, 6, 0, 0 );
   create_box((char*)Vectorout, "out", 6, 0, 0, 1); OutB = boxes-1;
 
   //   
   // connecting boxes
   //
   
     
   //-- vector in
   
   Boxlist[InB]->Yout->To[1]   = Boxlist[EndNeural+1]; // teacher
   Boxlist[InB]->Yout->Port[1] = 0;
   Boxlist[InB]->Yout->Size[1] = neurons[layers-1];
   if (!(Boxlist[InB]->Yout->Data[1] = new signal[neurons[layers-1]] ))
   {
      ErrOutput << MallocErr;
      exitus;
   }
   
   //-- end of actual NN
   
   Boxlist[EndNeural]->Yout->To[0] = Boxlist[EndNeural+1];
      
   //-- tester
   
   //---- Xin
   Boxlist[EndNeural+1]->Xin->From[0] = Boxlist[InB];  // desired values
   Boxlist[EndNeural+1]->Xin->Port[0] = 1;
   Boxlist[EndNeural+1]->Xin->Size[0] = neurons[layers-1];
   Boxlist[EndNeural+1]->Xin->Data[0] = Boxlist[InB]->Yout->Data[1];
   
   Boxlist[EndNeural+1]->Xin->From[1] = Boxlist[EndNeural];  // actual values
   Boxlist[EndNeural+1]->Xin->Port[1] = 0;
   Boxlist[EndNeural+1]->Xin->Size[1] = neurons[layers-1];
   Boxlist[EndNeural+1]->Xin->Data[1] = Boxlist[EndNeural]->Yout->Data[0];
   
   Boxlist[EndNeural+1]->Xin->From[2] = Boxlist[InB];  // input
   Boxlist[EndNeural+1]->Xin->Port[2] = 0;
   Boxlist[EndNeural+1]->Xin->Size[2] = neurons[0];
   Boxlist[EndNeural+1]->Xin->Data[2] = Boxlist[InB]->Yout->Data[0];
   
   //---- Yout
   int i;
   for (i=desired; i<=input; i++)
   {
      
      Boxlist[EndNeural+1]->Yout->To[i]   = Boxlist[OutB]; 
      Boxlist[EndNeural+1]->Yout->Port[i] = i;
       
      int size;
      
      switch (i)
      {
	 case desired:
	 case computed:
	 case error:     size = neurons[layers-1];
	                 break;
	 case distance:
	 case tdistance: size = 1;
	                 break;
	 case input:     size = neurons[0];
      }
	       
      Boxlist[EndNeural+1]->Yout->Size[i] = size;
      
      if (!(Boxlist[EndNeural+1]->Yout->Data[i] = new signal[size] ))
      {
         ErrOutput << MallocErr;
         exitus;
      }

   }
   
   //-- vector out
  
   Boxlist[OutB]->Xin->N = 0;
   
   for (i=desired; i<=input; i++)
   {
      
     Boxlist[OutB]->Xin->From[i] = Boxlist[EndNeural+1];
     Boxlist[OutB]->Xin->Port[i] = output[i];
         
     int size;
     switch (output[i])
     {
	case desired:
	case computed:
	case error:     size=neurons[layers-1];
	                Boxlist[OutB]->Xin->N++;
	                break;
	case distance:
	case tdistance: size=1;
	                Boxlist[OutB]->Xin->N++;
	                break;
	case input:     size = neurons[0];
	                Boxlist[OutB]->Xin->N++;
	                break;
	default:        size = 0;
     }
	   
   
     Boxlist[OutB]->Xin->Size[i] = size;
     Boxlist[OutB]->Xin->Data[i] = Boxlist[EndNeural+1]->Yout->Data[output[i]];
      
   } 
      
      
   Boxlist[OutB]->Xout->To[0]   = Boxlist[InB];
   Boxlist[OutB]->Xout->Port[0] = 0;
   Boxlist[OutB]->Xout->Size[0] = 0;
   

}
   

   
//// kernel::config_conf ///////////////////////////
//

// Sets up the boxes specific to the confusion subprogram,
// in this case a maximiser and a confusion box

void kernel::config_conf()
{
   
   create_box((char*)Maxim,"maxim", 1, 1, 0, 0 );
   create_box((char*)Confusion, "confusion", 2, 0, 0, 1); OutB = boxes-1;
   
   //   
   // connecting boxes
   //
   
  
   //-- vector in
   
   Boxlist[InB]->Yout->To[1]   = Boxlist[OutB]; // confusion
   Boxlist[InB]->Yout->Port[1] = 0;
   Boxlist[InB]->Yout->Size[1] = 1;
   if (!(Boxlist[InB]->Yout->Data[1] = new signal[1] ))
   {
      ErrOutput << MallocErr;
      exitus;
   }
 
   //-- end of actual NN
  
   Boxlist[EndNeural]->Yout->To[0] = Boxlist[EndNeural+1]; // maxim
      
   //-- maxim
   
   Boxlist[EndNeural+1]->Xin->From[0] = Boxlist[EndNeural];
   Boxlist[EndNeural+1]->Xin->Port[0] = 0;
   Boxlist[EndNeural+1]->Xin->Size[0] = neurons[layers-1];
   Boxlist[EndNeural+1]->Xin->Data[0] = Boxlist[EndNeural]->Yout->Data[0];

   Boxlist[EndNeural+1]->Yout->To[0]   = Boxlist[OutB];
   Boxlist[EndNeural+1]->Yout->Port[0] = 1;
   Boxlist[EndNeural+1]->Yout->Size[0] = 1;
   if (!(Boxlist[EndNeural+1]->Yout->Data[0] = new signal[1] ))
   {
      ErrOutput << MallocErr;
      exitus;
   }
   
   //-- confusion
   
   Boxlist[OutB]->Xin->From[0] = Boxlist[InB];
   Boxlist[OutB]->Xin->Port[0] = 1;
   Boxlist[OutB]->Xin->Size[0] = 1;
   Boxlist[OutB]->Xin->Data[0] = Boxlist[InB]->Yout->Data[1];
   
   Boxlist[OutB]->Xin->From[1] = Boxlist[EndNeural+1]; //maxim
   Boxlist[OutB]->Xin->Port[1] = 0;
   Boxlist[OutB]->Xin->Size[1] = 1;
   Boxlist[OutB]->Xin->Data[1] = Boxlist[EndNeural+1]->Yout->Data[0];
   
   Boxlist[OutB]->Xout->To[0]   = Boxlist[InB];
   Boxlist[OutB]->Xout->Port[0] = 0;
   Boxlist[OutB]->Xout->Size[0] = 0;
   
   create_box((char*)VectoroutDummy, "outdummy", 0, 0, 0, 0);
   // to allow for saving of the vector out parameters
   
 
}

//// kernel::config_class //////////////////////////
//

// Sets up the boxes specific to the classification subprogram,
// in this case a maximiser and a vector out.

void kernel::config_class()
{
   
   int size=(IndexOnly ? Nbest : 2*Nbest);
   create_box((char*)Maxim,"maxim", 1, size, 0, 0); 
   create_box((char*)Vectorout, "out", 1, 0, 0, 1); OutB = boxes-1;
   
   
   //   
   // connecting boxes
   //
   

   //-- end of actual NN
  
   Boxlist[EndNeural]->Yout->To[0] = Boxlist[EndNeural+1]; // maxim
      
   //-- maxim
   
   Boxlist[EndNeural+1]->Xin->From[0] = Boxlist[EndNeural];
   Boxlist[EndNeural+1]->Xin->Port[0] = 0;
   Boxlist[EndNeural+1]->Xin->Size[0] = neurons[layers-1];
   Boxlist[EndNeural+1]->Xin->Data[0] = Boxlist[EndNeural]->Yout->Data[0];


   Boxlist[EndNeural+1]->Yout->To[0]   = Boxlist[OutB];
   Boxlist[EndNeural+1]->Yout->Port[0] = 0;
   Boxlist[EndNeural+1]->Yout->Size[0] = size;
   if (!(Boxlist[EndNeural+1]->Yout->Data[0] = new signal[size] ))
   {
      ErrOutput << MallocErr;
      exitus;
   }
   
   //-- vector out
   
   Boxlist[OutB]->Xin->From[0] = Boxlist[EndNeural+1];
   Boxlist[OutB]->Xin->Port[0] = 0;
   Boxlist[OutB]->Xin->Size[0] = size;
   Boxlist[OutB]->Xin->Data[0] = Boxlist[EndNeural+1]->Yout->Data[0];
   
     
   Boxlist[OutB]->Xout->To[0]   = Boxlist[InB];
   Boxlist[OutB]->Xout->Port[0] = 0;
   Boxlist[OutB]->Xout->Size[0] = 0;

}

//// kernel::init_boxes ////////////////////////////
//

// Initialises the boxes that read their parameters from the kernel section

void kernel::init_boxes()
{

   FindSection(Kernel);
   
   if (Boxlist[InB]->init() != OKAY)
   {
      ErrOutput << Sign << "Initialisation of input block failed.\n";
      exitus;
   }
   
   if (Boxlist[OutB]->init() != OKAY)
   {
      ErrOutput << Sign << "Initialisation of output block failed.\n";
      exitus;
   }
     
   for (int i=EndNeural+1; i<OutB; i++)
   {
      if (Boxlist[i]->init() != OKAY)
      {
         ErrOutput << Sign << "Initialisation of block " 
	    << Boxlist[i]->Signature << " failed.\n";
         exitus;
       }
   }
   
   if ( (!(strcmp(Signature,TrainStr))) || (!(strcmp(Signature,ConfStr))) )
   {
      if (Boxlist[OutB+1]->init() != OKAY)
      {
	ErrOutput << Sign << "Initialisation of dummy block failed.\n";
	exitus;
      }
   }
     
       
}



//// kernel::process /////////////////////////////////
//

void kernel::process()
{

   blackBox *box;
   sysig    sig;
   
   box = Boxlist[InB];
   sig = box->forward();
   
   while (sig != NOsig)
   {
      switch (sig)
      {
         case Fsig:	box = box->Yout->To[0];
	                sig = box->forward();
			break;
			
	 case Bsig:	box=box->Xout->To[0];
			sig = box->backward();
			break;
			
	 case Errsig:	ErrOutput << Sign << box->Signature
	    			<< "'s method failed.\n";
      }
      
          
      if (strcmp(box->Signature,"in") == 0)
         if ( SaveCond() && (!strcmp(Signature,TrainStr)) )
	    if (save() != OKAY)
	        WarnOutput << Sign << "Saving into '"
	        	<< FileName << "' didn't end up okay.\n";
      
   }
}

//// kernel::shutdown /////////////////////////////
//

void kernel::shutdown()
{
   
   // shutdown all boxes
   int i;
   for(i=0; i<boxes; i++)
   { 
      if(Boxlist[i]->shutdown() != OKAY)
      {
	 ErrOutput << Sign << Boxlist[i]->Signature
	    << " didn't shutdown correct.\n";
	 exitus;
      }
   }
 
   // save at the end
   if (SaveEnable)
      if ( (!(strcmp(Signature,TrainStr))) || (!(strcmp(Signature,ConfStr))) )
	  if(save() != OKAY)
	     WarnOutput << Sign << "Final saving into '" 
		<< FileName << "' didn't end up okay.\n";
	  
}


//// kernel::ReadNextSection //////////////////////
//

// Would be better placed in blind.C, but since that's a rapid
// file, it was decided not to modify it.

result kernel::ReadNextSection(char *buffer)
{
   
   char c;
   int i=0;
   
   do
   {
      c=Getc();
   }
   while ((c!=EOF) && (c != '[') );
   
      
   if (c == EOF) 
   {
      return EMPTY;
   }
   else
   {
      IgnoreSpcLn();
      c=Getc();
      
      do
      {
	 buffer[i++]=c;
	 IgnoreSpcLn();
	 c=Getc();
      }
      while (c != ']');
      
      buffer[i]=0;
      SectionOffset=Ftell();
      return OKAY;
   }
      
         
}

//// kernel::save /////////////////////////////////
//

// Saves all into the specification file.
// Only called when training or calculating confusion tables.

result kernel::save()
{

   int i;
   
   // backup first
   
   if(BackupFName)
      if(Backup(BackupFName) != OKAY)
      {
         ErrOutput << Sign << "Backup into file '"
	    << BackupFName << "' failed.\n";
         return FAILED;
      }

   // section kernel
   
   WriteStr("[ "); WriteStr((char*)Kernel); WriteStr(" ]"); WriteLn();
   
   WriteVal((char*)TopologyKey,neurons[0]); WriteStr(" ");
   for (i=1; i<layers; i++) 
   {
      WriteVal(neurons[i]); WriteStr(" ");
   }
   WriteLn();
   
   WriteVal((char*)BckFileKey, BackupFName); WriteLn();
      
 
   if(SaveEnable)
      WriteVal((char*)SaveEnKey, (char*)TrueKey);
   else
      WriteVal((char*)SaveEnKey, (char*)FalseKey);
   WriteLn();
   
   WriteVal((char*)SaveCycKey, SaveCycle); WriteLn();
     
   WriteVal((char*)SeedKey, SeedVal); WriteLn();
   
   
   if (Boxlist[InB]->save() != OKAY) return FAILED;
   if (Boxlist[OutB]->save() != OKAY) return FAILED;
   
   WriteStr((char*)OutputKey); WriteStr(" = ");   
   for (i=0; i<5; i++)
   {
      if (output[i] != -1)
      {
	 WriteStr((char*)OutputKeys[output[i]]); WriteStr(" ");
      }
   }
   WriteLn();
   
   if (Boxlist[OutB+1]->save() != OKAY) return FAILED;
   
   if(IndexOnly)
      WriteVal((char*)IndexOnlyKey, (char*)TrueKey);
   else
      WriteVal((char*)IndexOnlyKey, (char*)FalseKey);
   WriteLn();
   
   WriteVal((char*)NBestKey, Nbest); WriteLn();
   

   for (i=StartNeural; i<=EndNeural; i++)
   {
      Boxlist[i]->WriteSection();
      if(Boxlist[i]->save() != OKAY)
	 WarnOutput << Sign 
	    << Boxlist[i]->Signature << " unable to save itself.\n";
   }
   

   // close the file

   Close();
   
   return OKAY;
   
}
      
//// kernel::SaveCond /////////////////////////////////
//

// test: are conditions for saving met now ?
	 
boolean kernel::SaveCond()
{
   if ( ( (!(strcmp(Signature,TrainStr))) || (!(strcmp(Signature,ConfStr))) ) 
      && ((SaveCycle-1) == SaveCntr++) && SaveCycle && SaveEnable ) 
         {
            SaveCntr = 0;
            return TRUE;
	 }
   else
      return FALSE;
      
}


//// kernel::create_box //////////////////////////////
//

// Determines whether or nor the BP version of the blocks has to be
// created.
// Note that confusion also creates BP blocks, because otherwise saving
// would result in loss of the LearnRate value

void kernel::create_box
   (char *Type, char *SignatStr, int xins,int youts,int yins,int xouts)

{
   
   blackBox *NewBox;
   
   if(!strcmp(Type, Vectorin))
      if ( (!(strcmp(Signature,IoStr))) )
      {
	 NewBox = new vectorin(Type, SignatStr, 0, 1, 1, 0);
      }
      else
      {
	 NewBox = new vectorin(Type, SignatStr, 0, 2, 1, 0);
      }
   
   else if(!strcmp(Type, Vectorout))
      NewBox = new vectorout(Type, SignatStr, xins, 0, 0, xouts);
   
   else if(!strcmp(Type, VectoroutDummy))
      NewBox = new vectoroutdummy(Type, SignatStr, xins, 0, 0, xouts);
   
   else if(!strcmp(Type, Maxim))
   {
      if (!(strcmp(Signature,ClassStr)))
      {
	 if (IndexOnly)
	 {
	    NewBox = new maxim((char*)Maxim, SignatStr, 1, Nbest, 0, 0);
	 }
	 else
	 {
	    NewBox = new maxim2((char*)Maxim2,SignatStr, 1, 2*Nbest, 0, 0);
	 }
      }
      if (!(strcmp(Signature,ConfStr)))
      {
	    NewBox = new maxim((char*)Maxim,SignatStr, 1, 1, 0, 0 );
      }
   }
   
   else if(!strcmp(Type, Logistic))
      if ( (!(strcmp(Signature,TrainStr))) || (!(strcmp(Signature,ConfStr))) ) 
      {
	 NewBox = new logisticBP((char*)LogisticBP, SignatStr, 1, 1, 1, 1);
      }
      else
      {
	 NewBox = new logistic((char*)Logistic, SignatStr, 1, 1, 0, 0);
      }
   
    else if(!strcmp(Type, BipSigmoid))
      if ( (!(strcmp(Signature,TrainStr))) || (!(strcmp(Signature,ConfStr))) )
      {
	 NewBox = new bipsigmoidBP((char*)BipSigmoidBP, SignatStr, 1, 1, 1, 1);
      }
      else
      {
	 NewBox = new bipsigmoid((char*)BipSigmoid, SignatStr, 1, 1, 0, 0);
      }
   
    else if(!strcmp(Type, SoftMax))
      if ( (!(strcmp(Signature,TrainStr))) || (!(strcmp(Signature,ConfStr))) )
      {
	 NewBox = new softmaxBP((char*)SoftMaxBP, SignatStr, 1, 1, 1, 1);
      }
      else
      {
	 NewBox = new softmax((char*)SoftMax, SignatStr, 1, 1, 0, 0);
      }
   
    else if(!strcmp(Type, Synapses))
      if ( (!(strcmp(Signature,TrainStr))) || (!(strcmp(Signature,ConfStr))) )
      {
	 NewBox = new synapsesBP((char*)SynapsesBP, SignatStr, 1, 1, 1, 1);
      }
      else
      {
	 NewBox = new synapses((char*)Synapses, SignatStr, 1, 1, 0, 0);
      }
      
    else if(!strcmp(Type, SynapsesTD))
      if ( (!(strcmp(Signature,TrainStr))) || (!(strcmp(Signature,ConfStr))) )
      {
	 NewBox = new synapsesTDBP((char*)SynapsesTDBP, SignatStr, 1, 1, 1, 1);
      }
      else
      {
	 NewBox = new synapsesTD((char*)SynapsesTD, SignatStr, 1, 1, 0, 0);
      }
   
    else if(!strcmp(Type, SynapsesTD2))
      if ( (!(strcmp(Signature,TrainStr))) || (!(strcmp(Signature,ConfStr))) )
      {
	 NewBox = new synapsesTDBP2((char*)SynapsesTDBP2, SignatStr, 1, 1, 1, 1);
      }
      else
      {
	 NewBox = new synapsesTD2((char*)SynapsesTD2, SignatStr, 1, 1, 0, 0);
      }
   
   else if(!strcmp(Type, Teacher))
      NewBox = new teacher(Type, SignatStr, xins, youts, yins, xouts );
   
   else if(!strcmp(Type, Tester))
      NewBox = new tester(Type, SignatStr, xins, youts, yins, xouts );


   else if(!strcmp(Type, Confusion))
      NewBox = new confusion(Type, SignatStr,2, 0, 0, 1);
   
   else if(!strcmp(Type, ConfusionDummy))
      NewBox = new confusiondummy(Type, SignatStr, 0, 0, 0, 0);
   
   else
   {
      ErrOutput << "Unknown box type: " << Type << '\n';
      exitus;
   }
   
   if (!(NewBox)) 
   {
    
      ErrOutput << MallocErr << SignatStr 
	<< " has not been created.\n";
      exitus;
   }
   
   if (boxes==MAXBOXES)
   {
      ErrOutput << Sign << "Too many boxes. Increase MAXBOXES ans recompile.\n";
      exitus;
   }
     
   Boxlist[boxes++] = NewBox;
   
}
