//////////////////////////////////////////////////////////////////////////////
//
// File:        mlpkrnl.h
// Directory:   /users/pbienst/mlp    
//
//////////////////////////////////////////////////////////////////////////////

#ifndef MLPKRNL_H
#define MLPKRNL_H

#define MAXBOXES 255
#define MAXLAYERS 255 // In general a layer consists of two boxes :
                      // a linear one (synapses) and a non linear one.
                      // However, zero or more than one non linear boxes
                      // are allowed, so the choice for MAXLAYERS is rather
                      // artificial

#include "blind.h"
#include "black.h"
#include "tools2.h"
#include "neural.h"
#include "neural2.h"

// subprogram keywords 
extern const char *IoStr;
extern const char *TrainStr;
extern const char *TestStr;
extern const char *ConfStr;
extern const char *ClassStr;

//system signals
extern const sysig 	NOsig, Fsig, Bsig, Errsig;

//kernels key names
extern const char *Kernel;
extern const char *TopologyKey;
extern const char *BckFileKey;
extern const char *SaveCycKey;
extern const char *SaveEnKey;
extern const char *TrueKey;
extern const char *FalseKey;
extern const char *IndexOnlyKey;
extern const char *NBestKey;

// constants used for output fields of trainer
typedef short int outp;
extern const outp desired,computed,error,distance,tdistance;
extern const char *OutputKey;
extern const char *OutputKeys[6];


   
//////////////////////////////////////////////////////////////////////////////
//
// CLASS: kernel
//
//////////////////////////////////////////////////////////////////////////////

class kernel : public blindBox
{
   
   protected:

      blackBox		*Boxlist[MAXBOXES];
      int               boxes; // number of created boxes 
      int               InB; // index of vector_in  
                             // first box is Boxlist[0]
                             // last box is Boxlist[boxes-1]
      int               StartNeural; // indices of start and end of the
      int               EndNeural;   // actual NN, i.e. the linear and non
	                             // linear layers
      int	        OutB; // Index of vector_out box
      
	 
      int               neurons[MAXLAYERS]; // no. of neurons in each layer
      int	 	layers;             // no. of layers
                              // first layer has neurons[0] neurons
                              // last layer has neurons[layers-1] neurons
      
      
      boolean		SaveEnable;
      char		*BackupFName;
      int		SaveCycle;	// zero value inhibits backup
      int 		SaveCntr;
      
      boolean		IndexOnly;
      int		Nbest;
      
      outp              output[6]; // contains the desired output fields 
                                   // for the trainer
      
      long		SeedVal;
           
   
   public:
      kernel(const char *SignStr, char *FName);
      void init_kernel();
      void config_general();
      void config_io();
      void config_train();
      void config_test();
      void config_conf();
      void config_class();
      void init_boxes();
      void process();
      void shutdown();

   protected:
      result ReadNextSection(char* buffer);
      result save();
      boolean SaveCond();	// true if save is to run
      void create_box
	 (char *Type, char *SignatStr, int xins,int youts,int yins,int xouts);

 
};


#endif
