//////////////////////////////////////////////////////////////////////////////
//
// File:	neural.C
// Directory:	/users/hartmann/rapid/src
//
//////////////////////////////////////////////////////////////////////////////


#include "neural.h"

// boxes--type identifiers
const char *LogisticBP0  = "logisticBP0";
const char *Logistic     = "logistic";
const char *LogisticBP   = "logisticBP";
const char *BipSigmoid	 = "bipsigmoid";
const char *BipSigmoidBP = "bipsigmoidBP";
const char *BipSigmoidRBP= "bipsigmoidRBP";
const char *BipSigmoidBBP= "bipsigmoidBBP";
const char *Synapses     = "synapses";
const char *SynapsesBP   = "synapsesBP";
const char *SynapsesMBP  = "synapsesMBP";
const char *SynapsesBBP  = "synapsesBBP";
const char *TeacherE     = "teacherE";
const char *TeacherWE    = "teacherWE";
const char *TeacherBE    = "teacherBE";
const char *Confusion    = "confusion";
const char *Energizer    = "energizer";
const char *EnergyE	 = "energyE";

// key words 
const char *MaximKey 	 = "Maximum";
const char *SlopeKey	 = "Slope";
const char *LRateKey	 = "LearnRate";
const char *LDecayKey	 = "LRateDecay";
const char *DecayKey 	 = "WeightDecay";
const char *WeightKey	 = "Weights";
const char *MaxRndKey	 = "RandMaximum";
const char *ThldKey 	 = "Thresholds";
const char *ConfTabKey	 = "ConfTable";
const char *ConfMatKey	 = "ConfMatrix";
const char *CodeSizeKey	 = "CodeRange";
const char *NKey	 = "N";
const char *BlSizeKey	 = "BlockSize";

// confusion table
const char *DesireKey	 = "desired:";
const char *TestKey	 = "tested:  ";
const char *CorrKey	 = "correct[%]:";
const char *TotalKey	 = "total:";
const char *TotalPKey	 = "total[%]:";
const char  ClassChar	 = '_';
const char  ClassThis 	 = '#';
const char *OverallKey	 = "performance[%]:";
const char *NbestKey	 = "N-best:";



//// logisticBP0::init /////////////////////////////////
//

result logisticBP0::init()
{
  
   // locate I/O vectors, set size
   X = Xin->Data[0];
   Y = Yout->Data[0];
   Xb = Xout->Data[0];
   Yb = Yin->Data[0];
   if( (N=Xin->Size[0]) != Yout->Size[0] )
   {
      ErrOutput << Sign <<"Forward input/output inconsistency.\n";
      return FAILED;
   }
   if( (N != Xin->Size[0]) || (N != Yout->Size[0]) )
   {
      ErrOutput << Sign <<"Backward input/output inconsistency.\n";
      return FAILED;
   }
    
   return OKAY;

};   



//// logisticBP0::forward /////////////////////////////////
//

sysig logisticBP0::forward()
{
   
   int i;
   
   // updating loop : logistic function
   for(i=0;i<N;i++)
      Y[i] = 1.0 / (1.0 + exp(-X[i]));
   
   
   return Fsig;

};   



//// logisticBP0::backward /////////////////////////////////
//

sysig logisticBP0::backward()
{
   
   int 		i;
   for(i=0;i<N;i++)
      Xb[i] = Y[i] * (1-Y[i]) * Yb[i];
   
   return Bsig;
   
};   



//// logistic::init /////////////////////////////////
//

result logistic::init()
{
  
   int i;
   
   
   // locate I/O vectors, set size
   X = Xin->Data[0];
   Y = Yout->Data[0];
   if( (N=Xin->Size[0]) != Yout->Size[0] )
   {
      ErrOutput << Sign <<"Forward inputs/outputs inconsistency.\n";
      return FAILED;
   }
   
   //create fields
   Maximum = Slope = NULL;
   AllocArray(Maximum, N);
   AllocArray(Slope, N);
   if(!Maximum || !Slope)
      return FAILED;
      
   //read MaximumVal, SlopeVal
   if(ReadVal((char*)MaximKey, MaximumVal) != OKAY)
      MaximumVal = 1.0;	// predefined 

   if(ReadVal((char*)SlopeKey, SlopeVal) != OKAY)
      SlopeVal = 1.0; 	// predefined


   //read fields Maximum, Slope if specified
   if(ReadArray((char*)MaximKey,Maximum,N) != OKAY)
      for(i=0;i<N;i++)
	 Maximum[i] = MaximumVal;

   if(ReadArray((char*)SlopeKey,Slope,N) != OKAY)
      for(i=0;i<N;i++)
	 Slope[i] = SlopeVal;
   
         
   return OKAY;

};   



//// logistic::save /////////////////////////////////
//

result logistic::save()
{
   
   WriteArray((char*)MaximKey, Maximum,N);
   WriteArray((char*)SlopeKey, Slope,N);

   return OKAY;

};   



//// logistic::forward /////////////////////////////////
//

sysig logistic::forward()
{
   
   int i;
   
   // updating loop : logistic function
   for(i=0;i<N;i++)
      Y[i] = Maximum[i] / (1.0+exp(-Slope[i]*X[i]));
   
   
   return Fsig;

};   



//// logistic::backward /////////////////////////////////
//

sysig logistic::backward()
{

   return Bsig;
   
}



//// logisticBP::init /////////////////////////////////
//

result logisticBP::init()
{

   // locate I/O vectors
   Xb = Xout->Data[0];
   Yb = Yin->Data[0];
   if( (Xout->Size[0] != Yin->Size[0]) || (Xout->Size[0] != Xin->Size[0]) )
   {
      ErrOutput << Sign << "Backward inputs/outputs inconsistency.\n";
      return FAILED;
   }

   // reading learning rate
   if(ReadVal((char*)LRateKey,LearnRate) != OKAY)
      LearnRate = 0.0;
   
   // the rest is the same as in logistic
   
   return logistic::init();
   
}



//// logisticBP::save /////////////////////////////////
//

result logisticBP::save()
{

   WriteVal((char*)LRateKey,LearnRate);WriteLn();
   // the rest is the same as in logistic

   return logistic::save();
   
}



//// logisticBP::backward /////////////////////////////////
//

sysig logisticBP::backward()
{
   
   int 		i;
   double	y;

   
   for(i=0;i<N;i++)
   {
      y = Y[i]/Maximum[i];
   
      // propagating (first) : the derivative of logistic function 
      //
      //	Xb[i] = Maximum[i] * Slope[i] * y * (1-y) * Yb[i];
      //
      
      Xb[i] = Slope[i] * Y[i] * (1-y) * Yb[i];
      
      
      if(LearnRate != 0.0)
         // adaptation
	 //
         //	Slope[i] += LearnRate * Maximum[i] * X[i] * y * (1-y) * Yb[i];
	 //
	 
	 Slope[i] += LearnRate * X[i] * Y[i] * (1-y) * Yb[i];
	 
   }
   
   return Bsig;
   
};   



//// bipsigmoid::init /////////////////////////////////
//

result bipsigmoid::init()
{
  
   int i;
   
   // locate I/O vectors, set size
   X = Xin->Data[0];
   Y = Yout->Data[0];
   if( (N=Xin->Size[0]) != Yout->Size[0] )
   {
      ErrOutput << Sign <<"Forward inputs/outputs inconsistency.\n";
      return FAILED;
   }
   
   //create fields
   Maximum = Slope = NULL;
   AllocArray(Maximum, N);
   AllocArray(Slope, N);
   if(!Maximum || !Slope)
      return FAILED;
      
   //read MaximumVal, SlopeVal
   if(ReadVal((char*)MaximKey, MaximumVal) != OKAY)
      MaximumVal = 1.0;	// predefined 

   if(ReadVal((char*)SlopeKey, SlopeVal) != OKAY)
      SlopeVal = 1.0; 	// predefined

   //read fields Maximum, Slope if specified
   if(ReadArray((char*)MaximKey,Maximum,N) != OKAY)
      for(i=0;i<N;i++)
	 Maximum[i] = MaximumVal;

   if(ReadArray((char*)SlopeKey,Slope,N) != OKAY)
      for(i=0;i<N;i++)
	 Slope[i] = SlopeVal;
   
         
   return OKAY;

};   



//// bipsigmoid::save /////////////////////////////////
//

result bipsigmoid::save()
{
   
   WriteArray((char*)MaximKey, Maximum,N);
   WriteArray((char*)SlopeKey, Slope,N);

   return OKAY;

};   



//// bipsigmoid::forward /////////////////////////////////
//

sysig bipsigmoid::forward()
{
   
   int i;
   
   // updating loop : bipsigmoid function
   for(i=0;i<N;i++)
      Y[i] = 2.0 * Maximum[i] / (1.0+exp(-Slope[i]*X[i])) - Maximum[i];

   return Fsig;

};   



//// bipsigmoidBP::init /////////////////////////////////
//

result bipsigmoidBP::init()
{

   // locate I/O vectors
   Xb = Xout->Data[0];
   Yb = Yin->Data[0];
   if( (Xout->Size[0] != Yin->Size[0]) || (Xout->Size[0] != Xin->Size[0]) )
   {
      ErrOutput << Sign << "Backward inputs/outputs inconsistency.\n";
      return FAILED;
   }

   // reading learning rate
   if(ReadVal((char*)LRateKey,LearnRate) != OKAY)
      LearnRate = 0.0;
   
   // the rest is the same as in bipsigmoid
   
   return bipsigmoid::init();
   
}



//// bipsigmoidBP::save /////////////////////////////////
//

result bipsigmoidBP::save()
{

   WriteVal((char*)LRateKey,LearnRate);WriteLn();
   // the rest is the same as in bipsigmoid

   return bipsigmoid::save();
   
}



//// bipsigmoidBP::backward /////////////////////////////////
//

sysig bipsigmoidBP::backward()
{
   
   int 		i;
   double 	y;

   
   // loop : the derivative of bipsigmoid function 
   for(i=0;i<N;i++)
   {
      y = Y[i]/Maximum[i]/2.0 + 0.5;
      
      // propagation (first !)
      Xb[i] = 2.0 * Maximum[i] * Slope[i] * y * (1-y) * Yb[i];
            
      // adaptation
      if(LearnRate != 0.0)
         Slope[i] += LearnRate * 2.0 * Maximum[i] * X[i] * y * (1-y) * Yb[i];
    
   }
   
   return Bsig;
   
};   



//// bipsigmoidRBP::backward /////////////////////////////////
//

sysig bipsigmoidRBP::backward()
{
   
   int 		i;
   double 	y;

   // loop : the derivative of bipsigmoid function 
   for(i=0;i<N;i++)
   {
      y = Y[i]/Maximum[i]/2.0 + 0.5;
      
      // propagation (first !)
      Xb[i] = Yb[i];
            
      // adaptation
      if(LearnRate != 0.0)
         Slope[i] += LearnRate * 2.0 * Maximum[i] * X[i] * y * (1-y) * Yb[i];
    
   }
   
   return Bsig;
   
};   



//// bipsigmoidBBP::init /////////////////////////////////
//

result bipsigmoidBBP::init()
{

   if((Yin->Size[1] != 1) || (Xout->Size[1] != 1))
   {
      ErrOutput << Sign 
	 << "Backward control signals must be scalars.\n";
      return FAILED;
   }
   
   CtrlIn = Yin->Data[1];
   CtrlOut= Xout->Data[1];
   
   return bipsigmoidBP::init();
   
};



//// bipsigmoidBBP::backward /////////////////////////////////
//

sysig bipsigmoidBBP::backward()
{
   
   if(*CtrlOut = *CtrlIn)

      return bipsigmoidBP::backward();   
    
   else
   
      return Bsig;
   
};   



//// synapses::init /////////////////////////////////
//

result synapses::init()
{
  
   // input/output locations and sizes
   X = Xin->Data[0];
   Nx = Xin->Size[0];
 
   V = Yout->Data[0];
   Nv = Yout->Size[0];

   
   // create weights and thresholds
   W = NULL;   Thld = NULL;
   AllocArray(W, Nv, Nx);
   AllocArray(Thld, Nv);
   if(!W || !Thld)
      return FAILED;
      
   // load weights and thresholds
   if(ReadArray((char*)WeightKey, W, Nv, Nx) != OKAY)
   {
      ErrOutput << Sign << "No 2-d weight matrix specified.\n";
      return FAILED;
   }

   if(ReadArray((char*)ThldKey, Thld, Nv) != OKAY)
   {
      ErrOutput << Sign << "No vector of thresholds specified.\n";
      return FAILED;
   }

 
   return OKAY;
   
};



//// synapses::save /////////////////////////////////
//

result synapses::save()
{
   
   WriteArray((char*)WeightKey, W, Nv, Nx);
   WriteArray((char*)ThldKey, Thld, Nv);

   return OKAY;

};   



//// synapses::forward /////////////////////////////////
//

sysig synapses::forward()
{

   int k,j;
   
   for(k=0;k<Nv;k++)
   {
      V[k]=0.0;
      for(j=0;j<Nx;j++)
	 V[k] += W[k][j] * X[j];
      V[k] -= Thld[k];
   }
   

   return Fsig;

};



//// synapses::backward /////////////////////////////////
//

sysig synapses::backward()
{

   return NOsig;

};




//// synapsesBP::init /////////////////////////////////
//

result synapsesBP::init()
{
   
   int k,j;

      
   // input/output locations and sizes
   X = Xin->Data[0];
   Xb = Xout->Data[0];
   Nx = Xin->Size[0];
   if(Xout->Size[0])
      if(Nx != Xout->Size[0])
      {
         ErrOutput << Sign << "Vectors X and Xb have diferent sizes.\n";
         return FAILED;
      }
 
   V = Yout->Data[0];
   Vb = Yin->Data[0];
   if((Nv = Yout->Size[0]) != Yin->Size[0])
   {
      ErrOutput << Sign << "Vectors V and Vb have diferent sizes.\n";
      return FAILED;
   }

   
   // reading learning rate
   if(ReadVal((char*)LRateKey,LearnRate) != OKAY)
   {
      ErrOutput << Sign << LRateKey << SepEQU << " not specified.\n";
      return FAILED;
   }
   
   
   // create weights and thresholds
   W = NULL;   Thld = NULL;
   AllocArray(W, Nv, Nx);
   AllocArray(Thld, Nv);
   if(!W || !Thld)
      return FAILED;

   // set weights, if not given then zeros
   boolean ZeroWeights = FALSE;
   if(ReadArray((char*)WeightKey, W, Nv, Nx) != OKAY)
   {
      for(k=0;k<Nv;k++)
         for(j=0;j<Nx;j++)
	    W[k][j] = 0.0;
      ZeroWeights = TRUE;
   }
   
   // set thresholds, if not given then zeros
   if(ReadArray((char*)ThldKey, Thld, Nv) != OKAY)
      for(k=0;k<Nv;k++)
     	 Thld[k] = 0.0;
   	
   // if some random bound is given
   //   add noise to the weights&thresholds
   // in this way the weights&thresholds can be preset in random manner
   if(ReadVal((char*)MaxRndKey, RandMax) == OKAY)
   { 
      // adding noise, DC = 0
      for(k=0;k<Nv;k++)
      {
	 for(j=0;j<Nx;j++)
	    W[k][j] = 2.0*RandMax * randdouble - RandMax;
     	 Thld[k] = 2.0*RandMax * randdouble - RandMax;
      }
   }
   else 
      if(ZeroWeights)
         WarnOutput << Sign << "All weights are set to zero.\n";
      

   return OKAY;
      
};




//// synapsesBP::save /////////////////////////////////
//

result synapsesBP::save()
{
   
   WriteVal((char*)LRateKey,LearnRate);WriteLn();
   WriteArray((char*)WeightKey, W, Nv, Nx);
   WriteArray((char*)ThldKey, Thld, Nv);

   return OKAY;

};   



//// synapsesBP::backward /////////////////////////////////
//

sysig synapsesBP::backward()
{

   int k,j;
   
   
   // backpropagation (first)
   if(Xout->Size[0])
     for(j=0;j<Nx;j++)
     {
	Xb[j] = 0.0;
	for(k=0;k<Nv;k++)
	   Xb[j] += W[k][j] * Vb[k];
     }
   
     
   // adaptation
   if(LearnRate != 0.0)
      for(k=0;k<Nv;k++)
      {
	 for(j=0;j<Nx;j++)
	    W[k][j] += LearnRate * Vb[k] * X[j];
         Thld[k] -= LearnRate * Vb[k];
      }
   

   return Bsig;

};



//// synapsesMBP::init /////////////////////////////////
//

result synapsesMBP::init()
{
   
   if(synapsesBP::init() == OKAY)
   {
      // previous time-step matrices
      ddW = NULL;	ddThld = NULL;
      AllocArray(ddW, Nv, Nx);  
      AllocArray(ddThld, Nv);
      if(!ddW || !ddThld)
         return FAILED;
      
      // set to zero
      int k,j;
      for(k=0;k<Nv;k++)
      {
	 for(j=0;j<Nx;j++)
	    ddW[k][j] = 0.0;
	 ddThld[k] = 0.0;
      }
      
      if(ReadVal((char*)DecayKey, Decay) != OKAY)
      {
	 ErrOutput << Sign << "Reading '" << DecayKey << "' failed.\n";
	 return FAILED;
      }
      
      return OKAY;
      
   }

   return FAILED;
      
};



//// synapsesMBP::save /////////////////////////////////
//

result synapsesMBP::save()
{
   
   WriteVal((char*)DecayKey, Decay);WriteLn();
   return synapsesBP::save();
      
};



//// synapsesMBP::backward /////////////////////////////////
//

sysig synapsesMBP::backward()
{

   int 		k,j;
   double	dW, dThld;
   
   // backpropagation (first)
   if(Xout->Size[0])
     for(j=0;j<Nx;j++)
     {
	Xb[j] = 0.0;
	for(k=0;k<Nv;k++)
	   Xb[j] += W[k][j] * Vb[k];
     }
   
     
   // adaptation
   if(LearnRate != 0.0)
      for(k=0;k<Nv;k++)
      {
	 for(j=0;j<Nx;j++)
	 {
	    dW = LearnRate * Vb[k] * X[j];
	    W[k][j] += dW + Decay*ddW[k][j];
	    ddW[k][j] = dW;
	 }
	 dThld = LearnRate * Vb[k];
         Thld[k] -= dThld - Decay*ddThld[k];
	 ddThld[k] = dThld;
      }
   

   return Bsig;

};



//// synapsesBBP::init /////////////////////////////////
//

result synapsesBBP::init()
{

   if((Yin->Size[1] != 1) || (Xout->Size[1] != 1))
   {
      ErrOutput << Sign 
	 << "Backward control signals are scalars.\n";
      return FAILED;
   }
   
   CtrlIn = Yin->Data[1];
   CtrlOut= Xout->Data[1];
   
   return synapsesBP::init();
   
};



//// synapsesBBP::backward /////////////////////////////////
//

sysig synapsesBBP::backward()
{
   
   if(*CtrlOut = *CtrlIn)
      
      return synapsesBP::backward();
    
   else
   
      return Bsig;
   
};



//// teacherE::init /////////////////////////////////
//

result teacherE::init()
{
  
   // input/output locations and sizes
   d = Xin->Data[0];
   X = Xin->Data[1];
   e = Xout->Data[0];
   if(Yout->N)
      dist = Yout->Data[0];
   else 
      dist = NULL;
   
   if(  ((N = Xin->Size[0]) != Xin->Size[1]) || (N != Xout->Size[0]) )
   {
      ErrOutput << Sign 
	 << "Vector lengths of desired and actual responses and "
	 << "of error don't match.\n";
      return FAILED;
   }

   if(dist)
      if(Yout->Size[0] != 1)
      {
         ErrOutput << Sign 
	    << "Distance is a scalar (not a vector).\n";
         return FAILED;
      }
      

   return OKAY;
   
};



//// teacherE::forward /////////////////////////////////
//

sysig teacherE::forward()
{

   int i;
  
   if(dist)
      *dist = SignalZero ;
   
   for(i=0;i<N;i++)
   {
      e[i] = d[i] - X[i];
      if(dist)
         *dist += e[i] * e[i];
   }
   
   if(dist)
      return Fsig;
   else
      return Bsig;

};



//// teacherE::backward /////////////////////////////////
//

sysig teacherE::backward()
{

   int i;
  
   // this is a bit luxury since error vector is effectively computed twice
   for(i=0;i<N;i++)
      e[i] = d[i] - X[i];
   

   return Bsig;

};



//// teacherWE::init /////////////////////////////////
//

result teacherWE::init()
{

   if(teacherE::init() != OKAY)
      return FAILED;

   k = NULL;
   AllocArray(k, N);
   if(!k)
      return FAILED;
   
   if(ReadArray((char*)WeightKey, k, N) != OKAY)
   {
      ErrOutput << Sign 
	 << "Reading '" << WeightKey << "' failed.\n";
      return FAILED;
   }
   
   return OKAY;

};



//// teacherWE::save /////////////////////////////////
//

result teacherWE::save()
{

   WriteArray((char*)WeightKey, k, N);   
   return OKAY;

};



//// teacherWE::forward /////////////////////////////////
//

sysig teacherWE::forward()
{

   int i,imax;
  
   for(imax=0,i=1;i<N;i++)
      if(d[i] > d[imax])
	 imax = i;
      
   if(dist)
      *dist = SignalZero ;      
      
   for(i=0;i<N;i++)
   {
      e[i] = k[imax]*(d[i] - X[i]);
      if(dist)
         *dist += e[i] * e[i];
   }

   if(dist)
      return Fsig;
   else
      return Bsig;

};



//// teacherWE::backward /////////////////////////////////
//

sysig teacherWE::backward()
{

   int i,imax;
  
   // this is a bit luxury since error vector is effectively computed twice
   for(imax=0,i=1;i<N;i++)
      if(d[i] > d[imax])
	 imax = i;

   for(i=0;i<N;i++)
      e[i] = k[imax]*(d[i] - X[i]);
   

   return Bsig;

};



//// teacherBE::init /////////////////////////////////
//

result teacherBE::init()
{ 
   
   // read BlockSize
   if(ReadVal((char*)BlSizeKey, BlockSize) != OKAY)
   {
      ErrOutput << Sign 
	 << "Unable to read '" << BlSizeKey << "'\n.";
      return FAILED;
   }
   
   // init counter;this will start initialization in forward() method
   C = -1; // something negativ !!!
   
   // check size of control signal
   CtrlDist = Xout->Data[1];
   if(Xout->Size[1] != 1)
   {
      ErrOutput << Sign 
	 << "Backward output control signal is a scalar value.\n";
      return FAILED;
   }
   
   return teacherE::init();
   
}


   
//// teacherBE::save /////////////////////////////////
//

result teacherBE::save()
{   
   WriteVal((char*)BlSizeKey, BlockSize);WriteLn();
   
   return teacherE::save();
 
}


   
//// teacherBE::forward /////////////////////////////////
//

sysig teacherBE::forward()
{

   int i;
   
   if(C < 0)
   {
      // clear IOs, counter
      for(i=0;i<N;i++)
	 e[i] = SignalZero;
      if(dist)
	 *dist = SignalZero;
      EDist = SignalZero;

      C=0;
   }
      
   // add error signal   
   for(i=0;i<N;i++)
   {
      e[i] += d[i] - X[i];
      EDist += e[i] * e[i];
   }
   
   // actualize CtrlDist and counter and go ahead
   if(++C < BlockSize)
   {
      *CtrlDist = SignalZero;
      return Bsig;
   }
   else
   {
      *CtrlDist = EDist;
      C = -1;

      if(dist)
      {
         *dist = EDist;
         return Fsig;
      }
      else
         return Bsig;
   }

};



//// teacherBE::backward /////////////////////////////////
//

sysig teacherBE::backward()
{

   return Bsig;

};



//// confusion::init /////////////////////////////////
//

result confusion::init()
{

   int r,c;
   
   
   // input/output locations and sizes
   if((Xin->Size[0] != 1) || (Xin->Size[1] != 1))
   {
      ErrOutput << Sign << "Input vector sizes are incorrect.\n";
      return FAILED;
   }
   
   // open files
   ConfTabName = ConfMatName = NULL;

   if(ReadVal((char*)ConfTabKey, ConfTabName) != OKAY)
   {
      ErrOutput << Sign << "Reading '" << ConfTabKey << "' failed.\n";
      return FAILED;
   }
   if(!(ConfTabFile = fopen(ConfTabName,"w")))
   {
      ErrOutput << Sign << FopenErr(ConfTabName);
      return FAILED;
   }
   
   if(ReadVal((char*)ConfMatKey, ConfMatName) != OKAY)
   {
      ErrOutput << Sign << "Reading '" << ConfMatKey << "' failed.\n";
      return FAILED;
   }
   if(!(ConfMatFile = fopen(ConfMatName,"w")))
   {
      ErrOutput << Sign << FopenErr(ConfMatName);
      return FAILED;
   }
   

   // read code size
   if(ReadVal((char*)CodeSizeKey, N) != OKAY)
   {
      ErrOutput << Sign << "Reading '" << CodeSizeKey << "' failed.\n";
      return FAILED;
   }

   // malloc matrix; zero init values
   M = NULL;
   AllocArray(M, N, N);
   if(!M)
      return FAILED;
   for(r=0;r<N;r++)
      for(c=0;c<N;c++)
	 M[r][c] = 0;
   
   return OKAY;
   
};



//// confusion::save /////////////////////////////////
//

result confusion::save()
{
   
   int r,c,s,gs;
   
   // first column width set to 15 chars
   // otherwise to 10 chars
   
   // header lines
   fprintf(ConfTabFile, "file:\t%s\n\n", FileName);
   fprintf(ConfTabFile, "%15s", DesireKey);
   for(c=0;c<N;c++)
      fprintf(ConfTabFile, "%9i%c", c, ClassChar);
   fprintf(ConfTabFile, "\n%15s\n", TestKey);
   
   // body
   for(r=gs=0; r<N; r++)
   {
      fprintf(ConfTabFile, "%13i%c:",r,ClassChar);
      for(c=0;c<N;c++)
      {
	 // into confusion table ...
	 fprintf(ConfTabFile, "%9i", M[r][c]);
	 if(r==c) 
	    fprintf(ConfTabFile, "%c",ClassThis);
	 else
	    fprintf(ConfTabFile, "%c",' ');
         gs += M[r][c];
	 
         // into confusion matrix ...
	 fprintf(ConfMatFile, "%10i", M[r][c]);
      }
      fprintf(ConfTabFile, "\n");
      fprintf(ConfMatFile, "\n");
   }
   
   // final total line
   fprintf(ConfTabFile, "\n%15s",TotalKey);
   for(c=0;c<N;c++)
   {
      for(r=s=0;r<N;r++)
	 s += M[r][c];
      fprintf(ConfTabFile, "%9i ", s);
   }   

   fprintf(ConfTabFile, "\n%15s",TotalPKey);
   for(c=0;c<N;c++)
   {
      for(r=s=0;r<N;r++)
	 s += M[r][c];
      fprintf(ConfTabFile, "%9.1f ", (100.0*s)/gs);
   }   

   fprintf(ConfTabFile, "\n%15s",CorrKey);
   for(c=0;c<N;c++)
   {
      for(r=s=0;r<N;r++)
	 s += M[r][c];
      fprintf(ConfTabFile, "%9.1f ", (100.0*M[c][c])/s);
   } 
   
   // 
   fprintf(ConfTabFile, "\n%15s\n%15s%9i\n","**** ",NbestKey,Xin->Size[1]);
   
   // overall rate line
   fprintf(ConfTabFile, "%15s",OverallKey);
   for(c=s=0;c<N;c++)
      s += M[c][c];
   fprintf(ConfTabFile, "%9.1f\n", (100.0*s)/gs);
      
   // fflush files
   fflush(ConfTabFile);
   fflush(ConfMatFile);
   
   // write file names and code range
   WriteVal((char*)ConfTabKey, ConfTabName); WriteLn();
   WriteVal((char*)ConfMatKey, ConfMatName); WriteLn();
   WriteVal((char*)CodeSizeKey, N); WriteLn();
      
 
   return OKAY;
   
};



//// confusion::forward /////////////////////////////////
//

sysig confusion::forward()
{

   // read input class indexes
   D = (int)*(Xin->Data[0]);
   if((D<0) || (D>=N))
   {
      ErrOutput << Sign 
	 << "Desired class index (" << D 
	 << ") out of range (" << N << ")\n";
      return FAILED;
   }

   X = (int)*(Xin->Data[1]);
   if((X<0) || (X>=N))
   {
      ErrOutput << Sign 
	 << "Tested class index (" << X 
	 << ") out of range (" << N << ")\n";
      return FAILED;
   }

   // increment element
   M[X][D] += 1;
   

   return Bsig;

};



//// energizer::init /////////////////////////////////
//

result energizer::init()
{

   S = 0.0;
   n = 0l;
   if(ReadVal((char*)NKey, N) != OKAY)
      N = (long) -1;
   
   if((Xin->Size[0] != 1) || (Yout->Size[0] != 1))
   {
      ErrOutput << Sign << "forward I/O vector sizes != 1.\n";
      return FAILED;
   }
   else
      return OKAY;
   
};



//// energizer::save /////////////////////////////////
//

result energizer::save()
{
   
   WriteVal((char*)NKey, N);
   WriteLn();

   return OKAY;
   
};
   


//// energizer::forward /////////////////////////////////
//

sysig energizer::forward()
{
   
   S += *(Xin->Data[0]);
   
   if(++n >= N)
   {
      if(N>0)
	 *(Yout->Data[0]) = S/N;
      else
	 *(Yout->Data[0]) = S;
      n = 0l;
      S = 0.0;

      return Fsig;
   }
   else
      
      return Bsig;
 
};



//// energizer::backward /////////////////////////////////
//

sysig energizer::backward()
{
   
   return Bsig;
    
};



//// energyE::init /////////////////////////////////
//

result energyE::init()
{
   
   if((N=Xin->Size[0]) != Xin->Size[1])
   {
      ErrOutput << Sign << "Input vectors are not matching.\n";
      return FAILED;
   }
   X = Xin->Data[0];
   Y = Xin->Data[1];
   E = SignalZero;
   
   return OKAY;

};



//// energyE::forward /////////////////////////////////
//

sysig energyE::forward()
{
   
   int 		i;
   signal 	D;
   for(i=0;i<N;i++)
   {
      D = (X[i] - Y[i]);
      E += D * D;
   }
   
   return Bsig;

};



//// energyE::shutdown /////////////////////////////////
//

result energyE::shutdown()
{
   
   fprintf(stdout,"%g\n",E);
   
   return OKAY;

};
