//////////////////////////////////////////////////////////////////////////////
//
// File:	tools2.C
// Directory:	/users/pbienst/mlp
//
//////////////////////////////////////////////////////////////////////////////

// modified subset of the rapid tools.C


#include "tools2.h"


// boxes--type identifiers

const char *Vectorin	   = "vector_in";
const char *Vectorout	   = "vector_out";
const char *VectoroutDummy = "vector_out_dummy";
const char *Maxim	   = "maxim";
const char *Maxim2         = "maxim2";


// keywords
const char *InFNameKey	   = "InputFile";
const char *OutFNameKey	   = "OutputFile";
const char *InFormatKey	   = "InputFormat";
const char *OutFormatKey   = "OutputFormat";
const char *AsciiKey	   = "ascii";
const char *BinKey	   = "binary";
const char *FillKey	   = "AppendZeros";
const char *MaxLenKey	   = "RandBufferSize";
const char *InTeeFNameKey  = "InputTeeFile";
const char *OutTeeFNameKey = "OutputTeeFile";
const char *OnKey	   = "ON";
const char *OffKey	   = "OFF";

// consts
const int ascii = 0;
const int binary = 1;


//// vectorin::vectorin /////////////////////////////////
//

vectorin::vectorin
   (char *TypeStr, char *SignatStr, int Xins, int Youts, int Yins, int Xouts)
   : blackBox(TypeStr, SignatStr, Xins, Youts, Yins, Xouts)
{

   FName = TeeFName = NULL;
   Format = ascii;
   FillZeros = Finished = FALSE;

};



//// vectorin::init /////////////////////////////////
//

result vectorin::init()
{

   result Res;


   /// reading file name
   if((Res=ReadVal((char*)InFNameKey, FName)) != OKAY)  // i.e. EMPTY
      FName = '\0';


   /// reading tee file name
   if((Res=ReadVal((char*)InTeeFNameKey, TeeFName)) != OKAY)  // i.e. EMPTY
      TeeFName = '\0';


   /// reading binary/ascii
   if((Res=FindIdentifier((char*)InFormatKey, SepEQU)) == OKAY)
   {
      IgnoreSpcLn();
      if(CheckStr((char*)AsciiKey) == TRUE)
	 Format = ascii;
      else if(CheckStr((char*)BinKey) == TRUE)
	 Format = binary;

      // testing if line is not empty, then error
      ReadVal();
      if(*Buffer)
      {
	 ErrOutput << Sign << InFormatKey << ": syntax error. " <<'\n';
	 return FAILED;
	 //otherwise ascii
      }
   }
   else
      if(Res != EMPTY)
	 return FAILED;
   // otherwise ascii


   /// reading boolean: append zeros or not at the end
   if((Res=FindIdentifier((char*)FillKey, SepEQU)) == OKAY)
   {
      IgnoreSpcLn();
      if(CheckStr((char*)TrueKey) == TRUE)
	 FillZeros = TRUE;
      else if(CheckStr((char*)FalseKey) == TRUE)
	 FillZeros = FALSE;

      // testing if line is not empty, then error
      ReadVal();
      if(*Buffer)
      {
	 ErrOutput << Sign << FillKey << ": syntax error. " <<'\n';
	 return FAILED;
	 //otherwise FillZeros = FALSE
      }

   }
   else
      if(Res != EMPTY)
	 return FAILED;
      // otherwise FillZeros = FALSE


   /// open input stream
   if(FName)
      InFile = fopen(FName,"r");
   else
      InFile = stdin;


   if(!InFile)   // testing fopen result
   {
      ErrOutput << Sign << FopenErr(FName);
      return FAILED;
   }


   /// open teefile
   if(TeeFName)
      if(!(TeeFile=fopen(TeeFName,"w")))
	 WarnOutput << Sign << FopenErr(TeeFName);


   return OKAY;

};



//// vectorin::save /////////////////////////////////
//

result vectorin::save()
{

   // file name
   WriteVal((char*)InFNameKey,FName);WriteLn();

   // format: ascii/binary
   switch (Format)
   {
      case ascii	:WriteVal((char*)InFormatKey, (char*)AsciiKey);
			 break;
      case binary	:WriteVal((char*)InFormatKey, (char*)BinKey);
   }
   WriteLn();

   // tee file
   if(TeeFName)
   {
      WriteVal((char*)InTeeFNameKey,TeeFName);
      WriteLn();
   }

   // append zeros: true/false
   switch(FillZeros)
   {
      case TRUE:	WriteVal((char*)FillKey, (char*)TrueKey);
			break;
      case FALSE:	WriteVal((char*)FillKey, (char*)FalseKey);
   }
   WriteLn();


   return OKAY;

};



//// vectorin::forward /////////////////////////////////
//

sysig vectorin::forward()
{

   int i,j;
   result Res;
   boolean StRead = FALSE;


   if(Finished)
      return NOsig;
      // i.e. EOF was reached in the past forward cycle

   for(i=0; i<Yout->N; i++)
      for(j=0; j<Yout->Size[i]; j++)
      {
	 switch(Format)
	 {
	    case ascii:
	       Res = fscanf(InFile,"%lf",Yout->Data[i]+j);
	       // testing error
	       if(Res == 0)
	       {
		  ErrOutput << Sign
		     << "Error while reading ascii from input stream.\n";
		  exitus;
	       }
	       else
		  break;
	    case binary:
	       Res = fread(Yout->Data[i]+j,sizeof(*(Yout->Data[i])),1,InFile);
	       // hopefully no error occured ...
	 }

	 // testing eof (-1 for ascii, 0 for binary mode)
	 if(Res != 1)
	 {
	    if(FillZeros) // filling by zeros allowed
	    {
	       if(StRead)
		  *(Yout->Data[i]+j) = SignalZero;
	       else   // eof reached and nothing was read yet
		  break;
	    }
	    else
	    {
	       if(StRead)
	       {
		  ErrOutput << Sign << "Unable to form an input vector.\n";
		  exitus;
	       }
	    }
	 }
	 else
	    StRead = TRUE;
      }


   //write to tee-file
   if(TeeFile && StRead)
      for(i=0; i<Yout->N; i++)
      {
	 for(j=0; j<Yout->Size[i]; j++)
	 {
	    fprintf(TeeFile, SignalFormat, *(Yout->Data[i]+j));
	    fprintf(TeeFile, "\t");
	 }

	 fprintf(TeeFile, "\n");
	 fflush(TeeFile);
      }

   if(Res !=1)
      Finished = TRUE;

   if(StRead)
      return Fsig;
   else
      return NOsig;

};



//// vectorin::backward /////////////////////////////////
//

sysig vectorin::backward()
{

   return forward();

};



//// vectorout::vectorout /////////////////////////////////
//

vectorout::vectorout
   (char *TypeStr, char *SignatStr, int Xins, int Youts, int Yins, int Xouts)
   :blackBox(TypeStr, SignatStr, Xins, Youts, Yins, Xouts)
{
   
   FName = TeeFName = NULL; 
   Format = ascii; 

};



//// vectorout::init /////////////////////////////////
//

result vectorout::init()
{
   
   result Res;

   /// reading file name
   if((Res=ReadVal((char*)OutFNameKey, FName)) != OKAY)  // i.e. EMPTY
      FName = NULL;


   /// reading tee file name
   if((Res=ReadVal((char*)OutTeeFNameKey, TeeFName)) != OKAY)  // i.e. EMPTY
      TeeFName = NULL;


   /// reading binary/ascii
   if((Res=FindIdentifier((char*)OutFormatKey, SepEQU)) == OKAY)
   {
      IgnoreSpcLn();
      if(CheckStr((char*)AsciiKey) == TRUE)
	 Format = ascii;
      else if(CheckStr((char*)BinKey) == TRUE)
	 Format = binary;

      // testing if line is not empty, then error
      ReadVal();
      if(*Buffer)
      {
	 ErrOutput << Sign << OutFormatKey << ": syntax error. " <<'\n';
	 return FAILED;
	 //otherwise ascii
      }
   }
   else
      if(Res != EMPTY)
	 return FAILED;
   // otherwise ascii


   /// open output stream
   if(FName)
      OutFile = fopen(FName,"w");
   else
      OutFile = stdout;


   if(!OutFile)   // testing fopen result
   {
      ErrOutput << Sign << FopenErr(FName);
      return FAILED;
   }


   /// open teefile
   if(TeeFName)
      if(!(TeeFile=fopen(TeeFName,"w")))
	 WarnOutput << Sign << "Unable to open tee-file '" << TeeFName << "'\n";


   return OKAY;

};



//// vectorout::save /////////////////////////////////
//

result vectorout::save()
{

   // file name
   WriteVal((char*)OutFNameKey,FName);WriteLn();

   // format: ascii/binary
   switch (Format)
   {
      case ascii	:WriteVal((char*)OutFormatKey, (char*)AsciiKey);
			 break;
      case binary	:WriteVal((char*)OutFormatKey, (char*)BinKey);
   }
   WriteLn();

   // tee file
   if(TeeFName)
   {
      WriteVal((char*)OutTeeFNameKey,TeeFName);WriteLn();
   }

   return OKAY;

};



//// vectorout::forward /////////////////////////////////
//

sysig vectorout::forward()
{

   int i,j;

   // no errors during fprintf | fwrite are assumed

   for(i=0; i<Xin->N; i++)
   {
      for(j=0; j<Xin->Size[i]; j++)
      {
	 switch(Format)
	 {
	    case ascii:
	       fprintf(OutFile,SignalFormat,(Xin->Data[i])[j]);
	       fprintf(OutFile,"\t");
	       break;
	    case binary:
	       fwrite(Xin->Data[i]+j,sizeof(*(Xin->Data[i])),1,OutFile);
	 }

	 //write to tee-file
	 if(TeeFile)
	 {
	    fprintf(TeeFile, SignalFormat, (Xin->Data[i])[j]);
	    fprintf(TeeFile, "\t");
	 }

      }

      if(Format == ascii)
	    fprintf(OutFile, "\n");

      if(TeeFile)
      {
	 fprintf(TeeFile, "\n");
	 fflush(TeeFile);
      }
   }

   fflush(OutFile);

   return Bsig;

};

//// vectoroutdummy::vectoroutdummy /////////////////////////////////
//

vectoroutdummy::vectoroutdummy
   (char *TypeStr, char *SignatStr, int Xins, int Youts, int Yins, int Xouts)
   :blackBox(TypeStr, SignatStr, Xins, Youts, Yins, Xouts)
{

   FName = TeeFName = NULL;
   Format = ascii;

};



//// vectoroutdummy::init /////////////////////////////////
//

result vectoroutdummy::init()
{

   result Res;

   /// reading file name
   FName = NULL;
   if((Res=ReadVal((char*)OutFNameKey, FName)) != OKAY)  // i.e. EMPTY
      FName = NULL;


   /// reading tee file name
   TeeFName = NULL;
   if((Res=ReadVal((char*)OutTeeFNameKey, TeeFName)) != OKAY)  // i.e. EMPTY
      TeeFName = NULL;


   /// reading binary/ascii
   if((Res=FindIdentifier((char*)OutFormatKey, SepEQU)) == OKAY)
   {
      IgnoreSpcLn();
      if(CheckStr((char*)AsciiKey) == TRUE)
	 Format = ascii;
      else if(CheckStr((char*)BinKey) == TRUE)
	 Format = binary;

      // testing if line is not empty, then error
      ReadVal();
      if(*Buffer)
      {
	 ErrOutput << Sign << OutFormatKey << ": syntax error. " <<'\n';
	 return FAILED;
	 //otherwise ascii
      }
   }
   else
      if(Res != EMPTY)
	 return FAILED;
   // otherwise ascii

   return OKAY;

};


//// vectoroutdummy::save /////////////////////////////////
//

result vectoroutdummy::save()
{

   // file name
   WriteVal((char*)OutFNameKey,FName);WriteLn();

   // format: ascii/binary
   switch (Format)
   {
      case ascii	:WriteVal((char*)OutFormatKey, (char*)AsciiKey);
			 break;
      case binary	:WriteVal((char*)OutFormatKey, (char*)BinKey);
   }
   WriteLn();

   // tee file
   if(TeeFName)
   {
      WriteVal((char*)OutTeeFNameKey,TeeFName);WriteLn();
   }

   return OKAY;

};


//// maxim::init /////////////////////////////////
//

result maxim::init()
{

   Nin = Xin->Size[0];
   Nout = Yout->Size[0];
   Input = Xin->Data[0];
   Output = Yout->Data[0];

   return OKAY;

};

//// maxim::forward /////////////////////////////////
//

sysig maxim::forward()
{

   int n,i,maxi,j;

   for(n=0;n<Nout;n++)
   {
      for(i=0,maxi=-1; i<Nin; i++)
      {
	 for(j=0; j<n; j++)
	    if(Output[j] == i)
	       break;
	 if(j==n)
	 {
	    if(maxi<0)
	       maxi=i;
	    else
	       if(Input[i] > Input[maxi])
		  maxi=i;
	 }
      }

      Output[n] = maxi;
   }

   return Fsig;
   
};



//// maxim2::init /////////////////////////////////
//

// maxim2 returns in the second half of the vector the actual values
// of the input as well

result maxim2::init()
{
   
   Nin = Xin->Size[0];
   if (Yout->Size[0]%2)
   {
      ErrOutput << Sign << "length of output vector must be even\n";
      return FAILED;
   }
   Nout = Yout->Size[0]/2;
   Input = Xin->Data[0];
   Output = Yout->Data[0];

   return OKAY;

};



//// maxim2::forward /////////////////////////////////
//

sysig maxim2::forward()
{
   
   int n,i,maxi,j;
   
   for(n=0;n<Nout;n++)
   {
      for(i=0,maxi=-1; i<Nin; i++)
      {
	 for(j=0; j<n; j++)
	    if(Output[j] == i)
	       break;
	 if(j==n)
	 {
	    if(maxi<0)
	       maxi=i;
	    else
	       if(Input[i] > Input[maxi])
		  maxi=i;
	 }
      }
   
      Output[n] = maxi;
   }
   
   for(n=Nout; n<2*Nout; n++)
   {
      Output[n]=Input[(int)Output[n-Nout]];
   }
      
   return Fsig;
   
};

