//////////////////////////////////////////////////////////////////////////////
//
// File:	tools2.h
// Directory:	/users/pbienst/mlp
//
//////////////////////////////////////////////////////////////////////////////

// modified subset of the rapid tools.h


#ifndef TOOLS2_H
#define TOOLS2_H

#include "black.h"

// boxes--type identifiers in tools2.C
extern const char *Vectorin;
extern const char *Vectorout;
extern const char *VectoroutDummy;
extern const char *Maxim;
extern const char *Maxim2;

// constants--defined in tools2.C
extern const int ascii;
extern const int binary;

extern const char *InFormatKey;
extern const char *OutFormatKey;
extern const char *InTeeFNameKey;
extern const char *OutTeeFNameKey;
extern const char *OnKey;
extern const char *OffKey;
extern const char *TrueKey;
extern const char *FalseKey;




//////////////////////////////////////////////////////////////////////////////
//
// CLASS: vectorin
//
//////////////////////////////////////////////////////////////////////////////

// it reads the input file or stdin into output-forward vectors
// -in ascii or binary
// -possibly makes the ascii tee file
// -unlimited number of input-forward and out/input-backward connections
//  allowed
// -possibly appends zeros to form the vector
// -remind: backward track is revearted

class vectorin : public blackBox
{
   
   protected:
      char 	*FName;
      FILE	*InFile;
      char	*TeeFName;
      FILE	*TeeFile;
      int 	Format;
      boolean	FillZeros;
      boolean   Finished;
      
   public:
      vectorin
         (char *TypeStr, char *SignatStr,int Xins,int Youts,int Yins,int Xouts);
	 
   protected:      
      virtual result init();
      virtual result save();
      
      virtual sysig forward();
      virtual sysig backward();
      

};



//////////////////////////////////////////////////////////////////////////////
//
// CLASS: vectorout
//
//////////////////////////////////////////////////////////////////////////////

// it writes all the input-forward vectors into file or stdout
//  -in ascii or binary
//  -possibly makes the ascii tee file
//  -unlimited number of output-forward and out/input-backward connections
//   allowed
//  -remind: forward track is revearted

class vectorout : public blackBox
{
   
   protected:
      char 	*FName;
      FILE	*OutFile;
      char	*TeeFName;
      FILE	*TeeFile;
      int 	Format;
            
   public:
      vectorout
         (char *TypeStr,char *SignatStr,int Xins,int Youts,int Yins,int Xouts);
	 
   protected:      
      virtual result init();
      virtual result save();
      
      virtual sysig forward();

};

//////////////////////////////////////////////////////////////////////////////
//
// CLASS: vectoroutdummy
//
//////////////////////////////////////////////////////////////////////////////

// just reads in the nessesay variables and save them afterwards

class vectoroutdummy : public blackBox
{
   
   protected:
      char 	*FName;
      FILE	*OutFile;
      char	*TeeFName;
      FILE	*TeeFile;
      int 	Format;
            
   public:
      vectoroutdummy
         (char *TypeStr,char *SignatStr,int Xins,int Youts,int Yins,int Xouts);
	 
   protected:      
      virtual result init();
      virtual result save();
  
};

//////////////////////////////////////////////////////////////////////////////
//
// CLASS: maxim
//
//////////////////////////////////////////////////////////////////////////////

// receives the signal vector; 
//  indexes of the first Nout highest values are at the output

class maxim : public blackBox
{
   
   protected:
      int	Nin, Nout;
      signal	*Input, *Output;
      
   public:
      maxim(char *TypeStr,char *SignatStr,int Xins,int Youts,int Yins,int Xouts)
	: blackBox(TypeStr, SignatStr, Xins, Youts, Yins, Xouts) {};
	 
   protected:      
      virtual result init();
      virtual sysig forward();

};

//////////////////////////////////////////////////////////////////////////////
//
// CLASS: maxim2
//
//////////////////////////////////////////////////////////////////////////////

// receives the signal vector; 
// indexes of the first Nout/2 highest values are at the output, followed
// by the actual values of those signals

class maxim2 : public blackBox
{
   
   protected:
      int	Nin, Nout;
      signal	*Input, *Output;
      
   public:
      maxim2(char *TypeStr,char *SignatStr,int Xins,int Youts,int Yins,int Xouts)
	: blackBox(TypeStr, SignatStr, Xins, Youts, Yins, Xouts) {};
	 
   protected:      
      virtual result init();
      virtual sysig forward();

};


#endif
